import pytest
import pyarrow as pa
import pybnesian as pbn
from pybnesian import BayesianNetworkType, BayesianNetwork, ConditionalBayesianNetwork, GaussianNetwork,\
    SemiparametricBN, KDENetwork, DiscreteBN, LinearGaussianCPD, CKDE, DiscreteFactor
import pickle
import util_test

@pytest.fixture
def gaussian_bytes():
    gaussian = GaussianNetwork(["a", "b", "c", "d"], [("a", "b")])
    return pickle.dumps(gaussian)

@pytest.fixture
def spbn_bytes():
    spbn = SemiparametricBN(["a", "b", "c", "d"], [("a", "b")], [("b", pbn.CKDEType())])
    return pickle.dumps(spbn)

@pytest.fixture
def kde_bytes():
    kde = KDENetwork(["a", "b", "c", "d"], [("a", "b")])
    return pickle.dumps(kde)

@pytest.fixture
def discrete_bytes():
    discrete = DiscreteBN(["a", "b", "c", "d"], [("a", "b")])
    return pickle.dumps(discrete)

class MyRestrictedGaussianNetworkType(BayesianNetworkType):
    def __init__(self):
        BayesianNetworkType.__init__(self)

    def is_homogeneous(self):
        return True

    def default_node_type(self):
        return pbn.LinearGaussianCPDType()

    def can_have_arc(self, model, source, target):
        return "a" in source

    def new_bn(self, nodes):
        return NewBN(nodes)

    def new_cbn(self, nodes, interface_nodes):
        return ConditionalNewBN(nodes, interface_nodes)

    def __str__(self):
        return "MyRestrictedGaussianNetworkType"

@pytest.fixture
def genericbn_bytes():
    gen = BayesianNetwork(MyRestrictedGaussianNetworkType(), ["a", "b", "c", "d"], [("a", "b")])
    return pickle.dumps(gen)

class NewBN(BayesianNetwork):
    def __init__(self, variables, arcs=None):
        if arcs is None:
            BayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables)
        else:
            BayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, arcs)

@pytest.fixture
def newbn_bytes():
    new = NewBN(["a", "b", "c", "d"], [("a", "b")])
    return pickle.dumps(new)

class NonHomogeneousType(BayesianNetworkType):
    def __init__(self):
        BayesianNetworkType.__init__(self)
    
    def is_homogeneous(self):
        return False

    def data_default_node_type(self, dt):
        if dt.equals(pa.float64()) or dt.equals(pa.float32()):
            return [pbn.LinearGaussianCPDType()]
        else:
            raise ValueError("Data type not compatible with NonHomogeneousType")

    def new_bn(self, nodes):
        return OtherBN(nodes)

    def new_cbn(self, nodes, interface_nodes):
        return ConditionalOtherBN(nodes, interface_nodes)

    def __str__(self):
        return "NonHomogeneousType"


class OtherBN(BayesianNetwork):
    def __init__(self, variables, arcs=None, node_types=None):
        if arcs is None:
            if node_types is None:
                BayesianNetwork.__init__(self, NonHomogeneousType(), variables)
            else:
                BayesianNetwork.__init__(self, NonHomogeneousType(), variables, node_types)
        else:
            if node_types is None:
                BayesianNetwork.__init__(self, NonHomogeneousType(), variables, arcs)
            else:
                BayesianNetwork.__init__(self, NonHomogeneousType(), variables, arcs, node_types)

        self.extra_info = "extra"

    def __getstate_extra__(self):
        return self.extra_info

    def __setstate_extra__(self, t):
        self.extra_info = t

@pytest.fixture
def otherbn_bytes():
    other = OtherBN(["a", "b", "c", "d"], [("a", "b")], [("b", pbn.LinearGaussianCPDType()),
                                                         ("c", pbn.CKDEType()),
                                                         ("d", pbn.DiscreteFactorType())])
    return pickle.dumps(other)

def test_serialization_bn_model(gaussian_bytes, spbn_bytes, kde_bytes, discrete_bytes, genericbn_bytes, newbn_bytes, otherbn_bytes):
    loaded_g = pickle.loads(gaussian_bytes)
    assert set(loaded_g.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_g.arcs() == [("a", "b")]
    assert loaded_g.type() == pbn.GaussianNetworkType()

    loaded_s = pickle.loads(spbn_bytes)
    assert set(loaded_s.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_s.arcs() == [("a", "b")]
    assert loaded_s.type() == pbn.SemiparametricBNType()
    assert loaded_s.node_types() == {'a': pbn.UnknownFactorType(),
                                     'b': pbn.CKDEType(),
                                     'c': pbn.UnknownFactorType(),
                                     'd': pbn.UnknownFactorType()}

    loaded_k = pickle.loads(kde_bytes)
    assert set(loaded_k.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_k.arcs() == [("a", "b")]
    assert loaded_k.type() == pbn.KDENetworkType()

    loaded_d = pickle.loads(discrete_bytes)
    assert set(loaded_d.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_d.arcs() == [("a", "b")]
    assert loaded_d.type() == pbn.DiscreteBNType()

    loaded_gen = pickle.loads(genericbn_bytes)
    assert set(loaded_gen.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_gen.arcs() == [("a", "b")]
    assert loaded_gen.type() == MyRestrictedGaussianNetworkType()

    loaded_nn = pickle.loads(newbn_bytes)
    assert set(loaded_g.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_nn.arcs() == [("a", "b")]
    assert loaded_nn.type() == MyRestrictedGaussianNetworkType()

    loaded_o = pickle.loads(otherbn_bytes)
    assert set(loaded_g.nodes()) == set(["a", "b", "c", "d"])
    assert loaded_o.arcs() == [("a", "b")]
    assert loaded_o.type() == NonHomogeneousType()
    assert loaded_o.node_types() == {'a': pbn.UnknownFactorType(),
                                     'b': pbn.LinearGaussianCPDType(),
                                     'c': pbn.CKDEType(),
                                     'd': pbn.DiscreteFactorType()}
    assert loaded_o.extra_info == "extra"

    assert loaded_nn.type() != loaded_o.type()

@pytest.fixture
def gaussian_partial_fit_bytes():
    gaussian = GaussianNetwork(["a", "b", "c", "d"], [("a", "b")])
    lg = pbn.LinearGaussianCPD("b", ["a"], [1, 2], 2)
    gaussian.add_cpds([lg])
    gaussian.include_cpd = True
    return pickle.dumps(gaussian)

@pytest.fixture
def gaussian_fit_bytes():
    gaussian = GaussianNetwork(["a", "b", "c", "d"], [("a", "b")])
    lg_a = LinearGaussianCPD("a", [], [0], 0.5)
    lg_b = LinearGaussianCPD("b", ["a"], [1, 2], 2)
    lg_c = LinearGaussianCPD("c", [], [2], 1)
    lg_d = LinearGaussianCPD("d", [], [3], 1.5)
    gaussian.add_cpds([lg_a, lg_b, lg_c, lg_d])
    gaussian.include_cpd = True
    return pickle.dumps(gaussian)

@pytest.fixture
def other_partial_fit_bytes():
    other = OtherBN(["a", "b", "c", "d"], [("a", "b")], [("b", pbn.LinearGaussianCPDType()),
                                                         ("c", pbn.CKDEType()),
                                                         ("d", pbn.DiscreteFactorType())])
    lg = LinearGaussianCPD("b", ["a"], [1, 2], 2)
    other.add_cpds([lg])
    other.include_cpd = True
    return pickle.dumps(other)

@pytest.fixture
def other_fit_bytes():
    other = OtherBN(["a", "b", "c", "d"], [("a", "b")], [("b", pbn.LinearGaussianCPDType()),
                                                         ("c", pbn.CKDEType()),
                                                         ("d", pbn.DiscreteFactorType())])
    cpd_a = LinearGaussianCPD("a", [], [0], 0.5)
    cpd_b = LinearGaussianCPD("b", ["a"], [1, 2], 2)
    
    df_continuous = util_test.generate_normal_data_indep(100)
    cpd_c = CKDE("c", [])
    cpd_c.fit(df_continuous)

    df_discrete = util_test.generate_discrete_data_dependent(100)
    df_discrete.columns = df_discrete.columns.str.lower()
    cpd_d = DiscreteFactor("d", [])
    cpd_d.fit(df_discrete)
    
    other.add_cpds([cpd_a, cpd_b, cpd_c, cpd_d])

    other.include_cpd = True
    return pickle.dumps(other)

def test_serialization_fitted_bn(gaussian_partial_fit_bytes, gaussian_fit_bytes, other_partial_fit_bytes, other_fit_bytes):
    # ####################
    # Gaussian partial fit
    # ####################
    loaded_partial = pickle.loads(gaussian_partial_fit_bytes)
    assert not loaded_partial.fitted()
    cpd = loaded_partial.cpd("b")
    assert cpd.variable() == "b"
    assert cpd.evidence() == ["a"]
    assert list(cpd.beta) == [1, 2]
    assert cpd.variance == 2

    # ####################
    # Gaussian fit
    # ####################
    loaded_fitted = pickle.loads(gaussian_fit_bytes)
    assert loaded_fitted.fitted()

    cpd_a = loaded_fitted.cpd("a")
    assert cpd_a.variable() == "a"
    assert cpd_a.evidence() == []
    assert cpd_a.beta == [0]
    assert cpd_a.variance == 0.5

    cpd_b = loaded_fitted.cpd("b")
    assert cpd_b.variable() == "b"
    assert cpd_b.evidence() == ["a"]
    assert list(cpd_b.beta) == [1, 2]
    assert cpd_b.variance == 2

    cpd_c = loaded_fitted.cpd("c")
    assert cpd_c.variable() == "c"
    assert cpd_c.evidence() == []
    assert cpd_c.beta == [2]
    assert cpd_c.variance == 1
    
    cpd_d = loaded_fitted.cpd("d")
    assert cpd_d.variable() == "d"
    assert cpd_d.evidence() == []
    assert cpd_d.beta == [3]
    assert cpd_d.variance == 1.5

    # ####################
    # OtherBN homogeneous partial fit
    # ####################
    loaded_other = pickle.loads(other_partial_fit_bytes)
    assert not loaded_other.fitted()
    cpd = loaded_partial.cpd("b")
    assert cpd.variable() == "b"
    assert cpd.evidence() == ["a"]
    assert list(cpd.beta) == [1, 2]
    assert cpd.variance == 2

    # ####################
    # OtherBN homogeneous fit
    # ####################
    loaded_other_fitted = pickle.loads(other_fit_bytes)
    assert loaded_other_fitted.fitted()

    cpd_a = loaded_other_fitted.cpd("a")
    assert cpd_a.variable() == "a"
    assert cpd_a.evidence() == []
    assert cpd_a.beta == [0]
    assert cpd_a.variance == 0.5
    assert cpd_a.type() == pbn.LinearGaussianCPDType()

    cpd_b = loaded_other_fitted.cpd("b")
    assert cpd_b.variable() == "b"
    assert cpd_b.evidence() == ["a"]
    assert list(cpd_b.beta) == [1, 2]
    assert cpd_b.variance == 2
    assert cpd_b.type() == pbn.LinearGaussianCPDType()

    cpd_c = loaded_other_fitted.cpd("c")
    assert cpd_c.variable() == "c"
    assert cpd_c.evidence() == []
    assert cpd_c.fitted()
    assert cpd_c.num_instances() == 100
    assert cpd_c.type() == pbn.CKDEType()

    cpd_d = loaded_other_fitted.cpd("d")
    assert cpd_d.variable() == "d"
    assert cpd_d.evidence() == []
    assert cpd_d.fitted()
    assert cpd_d.type() == pbn.DiscreteFactorType()


# ##########################
# Conditional BN
# ##########################

@pytest.fixture
def cond_gaussian_bytes():
    gaussian = pbn.ConditionalGaussianNetwork(["c", "d"], ["a", "b"], [("a", "c")])
    return pickle.dumps(gaussian)

@pytest.fixture
def cond_spbn_bytes():
    spbn = pbn.ConditionalSemiparametricBN(["c", "d"], ["a", "b"], [("a", "c")], [("c", pbn.CKDEType())])
    return pickle.dumps(spbn)

@pytest.fixture
def cond_kde_bytes():
    kde = pbn.ConditionalKDENetwork(["c", "d"], ["a", "b"], [("a", "c")])
    return pickle.dumps(kde)

@pytest.fixture
def cond_discrete_bytes():
    discrete = pbn.ConditionalDiscreteBN(["c", "d"], ["a", "b"], [("a", "c")])
    return pickle.dumps(discrete)

@pytest.fixture
def cond_genericbn_bytes():
    gen = ConditionalBayesianNetwork(MyRestrictedGaussianNetworkType(), ["c", "d"], ["a", "b"], [("a", "c")])
    return pickle.dumps(gen)

class ConditionalNewBN(ConditionalBayesianNetwork):
    def __init__(self, variables, interface, arcs=None):
        if arcs is None:
            ConditionalBayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, interface)
        else:
            ConditionalBayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, interface, arcs)

@pytest.fixture
def cond_newbn_bytes():
    new = ConditionalNewBN(["c", "d"], ["a", "b"], [("a", "c")])
    return pickle.dumps(new)

class ConditionalOtherBN(ConditionalBayesianNetwork):
    def __init__(self, variables, interface, arcs=None, node_types=None):
        if arcs is None:
            if node_types is None:
                ConditionalBayesianNetwork.__init__(self, NonHomogeneousType(), variables, interface)
            else:
                ConditionalBayesianNetwork.__init__(self, NonHomogeneousType(), variables, interface, node_types)
        else:
            if node_types is None:
                ConditionalBayesianNetwork.__init__(self, NonHomogeneousType(), variables, interface, arcs)
            else:
                ConditionalBayesianNetwork.__init__(self, NonHomogeneousType(), variables, interface, arcs, node_types)

        self.extra_info = "extra"

    def __getstate_extra__(self):
        return self.extra_info

    def __setstate_extra__(self, t):
        self.extra_info = t

@pytest.fixture
def cond_otherbn_bytes():
    other = ConditionalOtherBN(["c", "d"], ["a", "b"], [("a", "c")], [("b", pbn.LinearGaussianCPDType()),
                                                                      ("c", pbn.CKDEType()),
                                                                      ("d", pbn.DiscreteFactorType())])
    return pickle.dumps(other)



def test_serialization_conditional_bn_model(cond_gaussian_bytes, cond_spbn_bytes, cond_kde_bytes, 
                                            cond_discrete_bytes, cond_genericbn_bytes,
                                            cond_newbn_bytes, cond_otherbn_bytes,
                                            newbn_bytes, otherbn_bytes):
    loaded_g = pickle.loads(cond_gaussian_bytes)
    assert set(loaded_g.nodes()) == set(["c", "d"])
    assert set(loaded_g.interface_nodes()) == set(["a", "b"])
    assert loaded_g.arcs() == [("a", "c")]
    assert loaded_g.type() == pbn.GaussianNetworkType()

    loaded_s = pickle.loads(cond_spbn_bytes)
    assert set(loaded_s.nodes()) == set(["c", "d"])
    assert set(loaded_s.interface_nodes()) == set(["a", "b"])
    assert loaded_s.arcs() == [("a", "c")]
    assert loaded_s.type() == pbn.SemiparametricBNType()
    assert loaded_s.node_types() == {'c': pbn.CKDEType(),
                                     'd': pbn.UnknownFactorType()}

    loaded_k = pickle.loads(cond_kde_bytes)
    assert set(loaded_k.nodes()) == set(["c", "d"])
    assert set(loaded_k.interface_nodes()) == set(["a", "b"])
    assert loaded_k.arcs() == [("a", "c")]
    assert loaded_k.type() == pbn.KDENetworkType()

    loaded_d = pickle.loads(cond_discrete_bytes)
    assert set(loaded_d.nodes()) == set(["c", "d"])
    assert set(loaded_d.interface_nodes()) == set(["a", "b"])
    assert loaded_d.arcs() == [("a", "c")]
    assert loaded_d.type() == pbn.DiscreteBNType()

    loaded_gen = pickle.loads(cond_genericbn_bytes)
    assert set(loaded_gen.nodes()) == set(["c", "d"])
    assert set(loaded_gen.interface_nodes()) == set(["a", "b"])
    assert loaded_gen.arcs() == [("a", "c")]
    assert loaded_gen.type() == MyRestrictedGaussianNetworkType()

    loaded_nn = pickle.loads(cond_newbn_bytes)
    assert set(loaded_nn.nodes()) == set(["c", "d"])
    assert set(loaded_nn.interface_nodes()) == set(["a", "b"])
    assert loaded_nn.arcs() == [("a", "c")]
    assert loaded_nn.type() == MyRestrictedGaussianNetworkType()

    loaded_o = pickle.loads(cond_otherbn_bytes)
    assert set(loaded_o.nodes()) == set(["c", "d"])
    assert set(loaded_o.interface_nodes()) == set(["a", "b"])
    assert loaded_o.arcs() == [("a", "c")]
    assert loaded_o.type() == NonHomogeneousType()
    assert loaded_o.node_types() == {'c': pbn.CKDEType(),
                                     'd': pbn.DiscreteFactorType()}
    assert loaded_o.extra_info == "extra"

    assert loaded_nn.type() != loaded_o.type()

    loaded_unconditional_nn = pickle.loads(newbn_bytes)
    loaded_unconditional_o = pickle.loads(otherbn_bytes)

    assert loaded_nn.type() == loaded_unconditional_nn.type()
    assert loaded_o.type() == loaded_unconditional_o.type()

@pytest.fixture
def cond_gaussian_partial_fit_bytes():
    gaussian = pbn.ConditionalGaussianNetwork(["c", "d"], ["a", "b"], [("a", "c")])
    lg = LinearGaussianCPD("c", ["a"], [1, 2], 2)
    gaussian.add_cpds([lg])
    gaussian.include_cpd = True
    return pickle.dumps(gaussian)

@pytest.fixture
def cond_gaussian_fit_bytes():
    gaussian = pbn.ConditionalGaussianNetwork(["c", "d"], ["a", "b"], [("a", "c")])
    lg_c = LinearGaussianCPD("c", ["a"], [1, 2], 2)
    lg_d = LinearGaussianCPD("d", [], [3], 1.5)
    gaussian.add_cpds([lg_c, lg_d])
    gaussian.include_cpd = True
    return pickle.dumps(gaussian)

@pytest.fixture
def cond_other_partial_fit_bytes():
    other = ConditionalOtherBN(["c", "d"], ["a", "b"], [("a", "c")], [("c", pbn.CKDEType()),
                                                                      ("d", pbn.LinearGaussianCPDType())])
    lg = LinearGaussianCPD("d", [], [3], 1.5)
    other.add_cpds([lg])
    other.include_cpd = True
    return pickle.dumps(other)

@pytest.fixture
def cond_other_fit_bytes():
    other = ConditionalOtherBN(["c", "d"], ["a", "b"], [("a", "c")], [("c", pbn.CKDEType()),
                                                                      ("d", pbn.DiscreteFactorType())])
    cpd_c = CKDE("c", ["a"])
    cpd_d = DiscreteFactor("d", [])
    
    df_continuous = util_test.generate_normal_data_indep(100)
    cpd_c.fit(df_continuous)

    df_discrete = util_test.generate_discrete_data_dependent(100)
    df_discrete.columns = df_discrete.columns.str.lower()
    cpd_d = DiscreteFactor("d", [])
    cpd_d.fit(df_discrete)

    other.add_cpds([cpd_c, cpd_d])
    
    other.include_cpd = True
    return pickle.dumps(other)

def test_serialization_fitted_conditional_bn(cond_gaussian_partial_fit_bytes, cond_gaussian_fit_bytes,
                                             cond_other_partial_fit_bytes, cond_other_fit_bytes):
    # ####################
    # Gaussian partial fit
    # ####################
    loaded_partial = pickle.loads(cond_gaussian_partial_fit_bytes)
    assert not loaded_partial.fitted()
    cpd = loaded_partial.cpd("c")
    assert cpd.variable() == "c"
    assert cpd.evidence() == ["a"]
    assert list(cpd.beta) == [1, 2]
    assert cpd.variance == 2

    # ####################
    # Gaussian fit
    # ####################
    loaded_fitted = pickle.loads(cond_gaussian_fit_bytes)
    assert loaded_fitted.fitted()

    cpd_c = loaded_fitted.cpd("c")
    assert cpd_c.variable() == "c"
    assert cpd_c.evidence() == ["a"]
    assert list(cpd_c.beta) == [1, 2]
    assert cpd_c.variance == 2
    
    cpd_d = loaded_fitted.cpd("d")
    assert cpd_d.variable() == "d"
    assert cpd_d.evidence() == []
    assert cpd_d.beta == [3]
    assert cpd_d.variance == 1.5

    # ####################
    # OtherBN homogeneous partial fit
    # ####################
    loaded_other = pickle.loads(cond_other_partial_fit_bytes)
    assert not loaded_other.fitted()
    cpd = loaded_other.cpd("d")
    assert cpd.variable() == "d"
    assert cpd.evidence() == []
    assert cpd.beta == [3]
    assert cpd.variance == 1.5

    # ####################
    # OtherBN homogeneous fit
    # ####################
    loaded_other_fitted = pickle.loads(cond_other_fit_bytes)
    assert loaded_other_fitted.fitted()

    cpd_c = loaded_other_fitted.cpd("c")
    assert cpd_c.variable() == "c"
    assert cpd_c.evidence() == ["a"]
    assert cpd_c.fitted()
    assert cpd_c.num_instances() == 100
    assert cpd_c.type() == pbn.CKDEType()

    cpd_d = loaded_other_fitted.cpd("d")
    assert cpd_d.variable() == "d"
    assert cpd_d.evidence() == []
    assert cpd_d.fitted()
    assert cpd_d.type() == pbn.DiscreteFactorType()

    assert loaded_other_fitted.extra_info == "extra"
    assert loaded_other.type() == loaded_other_fitted.type()

# ##########################
# Dynamic BN
# ##########################

@pytest.fixture
def dyn_gaussian_bytes():
    gaussian = pbn.DynamicGaussianNetwork(["a", "b", "c", "d"], 2)
    gaussian.static_bn().add_arc("a_t_2", "d_t_1")
    gaussian.transition_bn().add_arc("c_t_2", "b_t_0")
    return pickle.dumps(gaussian)

@pytest.fixture
def dyn_spbn_bytes():
    spbn = pbn.DynamicSemiparametricBN(["a", "b", "c", "d"], 2)
    spbn.static_bn().add_arc("a_t_2", "d_t_1")
    spbn.transition_bn().add_arc("c_t_2", "b_t_0")
    spbn.transition_bn().set_node_type("b_t_0", pbn.CKDEType())
    return pickle.dumps(spbn)

@pytest.fixture
def dyn_kde_bytes():
    kde = pbn.DynamicKDENetwork(["a", "b", "c", "d"], 2)
    kde.static_bn().add_arc("a_t_2", "d_t_1")
    kde.transition_bn().add_arc("c_t_2", "b_t_0")
    return pickle.dumps(kde)

@pytest.fixture
def dyn_discrete_bytes():
    discrete = pbn.DynamicDiscreteBN(["a", "b", "c", "d"], 2)
    discrete.static_bn().add_arc("a_t_2", "d_t_1")
    discrete.transition_bn().add_arc("c_t_2", "b_t_0")
    return pickle.dumps(discrete)

@pytest.fixture
def dyn_genericbn_bytes():
    gen = pbn.DynamicBayesianNetwork(MyRestrictedGaussianNetworkType(), ["a", "b", "c", "d"], 2)
    gen.static_bn().add_arc("a_t_2", "d_t_1")
    gen.transition_bn().add_arc("a_t_2", "b_t_0")
    return pickle.dumps(gen)

class DynamicNewBN(pbn.DynamicBayesianNetwork):
    def __init__(self, variables, markovian_order):
        pbn.DynamicBayesianNetwork.__init__(self, MyRestrictedGaussianNetworkType(), variables, markovian_order)

class DynamicOtherBN(pbn.DynamicBayesianNetwork):
    def __init__(self, variables, markovian_order, static_bn=None, transition_bn=None):
        if static_bn is None or transition_bn is None:
            pbn.DynamicBayesianNetwork.__init__(self, NonHomogeneousType(), variables, markovian_order)
        else:
            pbn.DynamicBayesianNetwork.__init__(self, variables, markovian_order, static_bn, transition_bn)
        self.extra_info = "extra"

    def __getstate_extra__(self):
        return self.extra_info

    def __setstate_extra__(self, t):
        self.extra_info = t

@pytest.fixture
def dyn_newbn_bytes():
    new = DynamicNewBN(["a", "b", "c", "d"], 2)
    new.static_bn().add_arc("a_t_2", "d_t_1")
    new.transition_bn().add_arc("a_t_2", "b_t_0")
    return pickle.dumps(new)

@pytest.fixture
def dyn_otherbn_bytes():
    other = DynamicOtherBN(["a", "b", "c", "d"], 2)
    other.static_bn().add_arc("a_t_2", "d_t_1")
    other.static_bn().set_node_type("c_t_1", pbn.DiscreteFactorType())
    other.static_bn().set_node_type("d_t_1", pbn.CKDEType())

    other.transition_bn().add_arc("a_t_2", "b_t_0")
    other.transition_bn().set_node_type("d_t_0", pbn.CKDEType())
    return pickle.dumps(other)

def test_serialization_dbn_model(dyn_gaussian_bytes, dyn_spbn_bytes, dyn_kde_bytes, dyn_discrete_bytes,
                                dyn_genericbn_bytes, dyn_newbn_bytes, dyn_otherbn_bytes):
    loaded_g = pickle.loads(dyn_gaussian_bytes)
    assert set(loaded_g.variables()) == set(["a", "b", "c", "d"])
    assert loaded_g.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_g.transition_bn().arcs() == [("c_t_2", "b_t_0")]
    assert loaded_g.type() == pbn.GaussianNetworkType()

    loaded_s = pickle.loads(dyn_spbn_bytes)
    assert set(loaded_s.variables()) == set(["a", "b", "c", "d"])
    assert loaded_s.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_s.transition_bn().arcs() == [("c_t_2", "b_t_0")]
    assert loaded_s.type() == pbn.SemiparametricBNType()
    node_types = {v + "_t_0": pbn.UnknownFactorType() for v in loaded_s.variables()}
    node_types["b_t_0"] = pbn.CKDEType()
    assert loaded_s.transition_bn().node_types() == node_types

    loaded_k = pickle.loads(dyn_kde_bytes)
    assert set(loaded_k.variables()) == set(["a", "b", "c", "d"])
    assert loaded_k.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_k.transition_bn().arcs() == [("c_t_2", "b_t_0")]
    assert loaded_k.type() == pbn.KDENetworkType()

    loaded_d = pickle.loads(dyn_discrete_bytes)
    assert set(loaded_d.variables()) == set(["a", "b", "c", "d"])
    assert loaded_d.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_d.transition_bn().arcs() == [("c_t_2", "b_t_0")]
    assert loaded_d.type() == pbn.DiscreteBNType()

    loaded_gen = pickle.loads(dyn_genericbn_bytes)
    assert set(loaded_gen.variables()) == set(["a", "b", "c", "d"])
    assert loaded_gen.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_gen.transition_bn().arcs() == [("a_t_2", "b_t_0")]
    assert loaded_gen.type() == MyRestrictedGaussianNetworkType()

    loaded_nn = pickle.loads(dyn_newbn_bytes)
    assert set(loaded_nn.variables()) == set(["a", "b", "c", "d"])
    assert loaded_nn.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_nn.transition_bn().arcs() == [("a_t_2", "b_t_0")]
    assert loaded_nn.type() == MyRestrictedGaussianNetworkType()

    loaded_other = pickle.loads(dyn_otherbn_bytes)
    assert set(loaded_other.variables()) == set(["a", "b", "c", "d"])
    assert loaded_other.static_bn().arcs() == [("a_t_2", "d_t_1")]
    assert loaded_other.transition_bn().arcs() == [("a_t_2", "b_t_0")]
    assert loaded_other.type() == NonHomogeneousType()
    assert loaded_other.extra_info == "extra"

    assert loaded_other.static_bn().node_type("c_t_1") == pbn.DiscreteFactorType()
    assert loaded_other.static_bn().node_type("d_t_1") == pbn.CKDEType()
    assert loaded_other.transition_bn().node_type("d_t_0") == pbn.CKDEType()

@pytest.fixture
def dyn_gaussian_partial_fit_bytes():
    gaussian = pbn.DynamicGaussianNetwork(["a", "b", "c", "d"], 2)
    gaussian.static_bn().add_arc("a_t_2", "d_t_1")
    gaussian.transition_bn().add_arc("c_t_2", "b_t_0")
    lg = LinearGaussianCPD("d_t_1", ["a_t_2"], [1, 2], 2)
    gaussian.static_bn().add_cpds([lg])
    lg = LinearGaussianCPD("b_t_0", ["c_t_2"], [3, 4], 5)
    gaussian.transition_bn().add_cpds([lg])
    gaussian.include_cpd = True
    return pickle.dumps(gaussian)

@pytest.fixture
def dyn_gaussian_fit_bytes():
    gaussian = pbn.DynamicGaussianNetwork(["a", "b", "c", "d"], 2)
    gaussian.static_bn().add_arc("a_t_2", "d_t_1")
    gaussian.transition_bn().add_arc("c_t_2", "b_t_0")
    df = util_test.generate_normal_data_indep(1000)
    gaussian.fit(df)
    gaussian.include_cpd = True
    return pickle.dumps(gaussian)

@pytest.fixture
def dyn_other_partial_fit_bytes():
    variables = ["a", "b", "c", "d"]
    static_nodes = [v + "_t_" + str(m) for v in variables for m in range(1, 3)]
    transition_nodes = [v + "_t_0" for v in variables]

    other_static = OtherBN(static_nodes, [("a_t_2", "d_t_1")], [("b_t_1", pbn.DiscreteFactorType()),
                                                                ("c_t_1", pbn.CKDEType()),
                                                                ("d_t_1", pbn.LinearGaussianCPDType())])
    lg = LinearGaussianCPD("d_t_1", ["a_t_2"], [1, 2], 2)
    other_static.add_cpds([lg])

    other_transition = ConditionalOtherBN(transition_nodes,
                                          static_nodes,
                                          [("a_t_2", "d_t_0")],
                                          [("b_t_0", pbn.DiscreteFactorType()),
                                            ("c_t_0", pbn.CKDEType()),
                                            ("d_t_0", pbn.LinearGaussianCPDType())])
    lg = LinearGaussianCPD("d_t_0", ["a_t_2"], [3, 4], 1.5)
    other_transition.add_cpds([lg])

    assert other_static.type() == other_transition.type()

    dyn_other = DynamicOtherBN(variables, 2, other_static, other_transition)
    dyn_other.include_cpd = True
    return pickle.dumps(dyn_other)

@pytest.fixture
def dyn_other_fit_bytes():
    variables = ["a", "b", "c", "d"]
    static_nodes = [v + "_t_" + str(m) for v in variables for m in range(1, 3)]
    transition_nodes = [v + "_t_0" for v in variables]

    other_static = OtherBN(static_nodes, [("a_t_2", "d_t_1")], [("b_t_2", pbn.DiscreteFactorType()),
                                                                ("b_t_1", pbn.DiscreteFactorType()),
                                                                ("c_t_1", pbn.CKDEType()),
                                                                ("d_t_1", pbn.LinearGaussianCPDType())])
    lg = LinearGaussianCPD("d_t_1", ["a_t_2"], [1, 2], 2)
    other_static.add_cpds([lg])

    other_transition = ConditionalOtherBN(transition_nodes,
                                          static_nodes,
                                          [("a_t_2", "d_t_0")],
                                          [("b_t_0", pbn.DiscreteFactorType()),
                                            ("c_t_0", pbn.CKDEType()),
                                            ("d_t_0", pbn.LinearGaussianCPDType())])
    lg = LinearGaussianCPD("d_t_0", ["a_t_2"], [3, 4], 1.5)
    other_transition.add_cpds([lg])

    assert other_static.type() == other_transition.type()

    dyn_other = DynamicOtherBN(variables, 2, other_static, other_transition)
    df_continuous = util_test.generate_normal_data_indep(1000)
    df_discrete = util_test.generate_discrete_data_dependent(1000)
    df = df_continuous
    df["b"] = df_discrete["B"]
    dyn_other.fit(df)
    dyn_other.include_cpd = True
    return pickle.dumps(dyn_other)

def test_serialization_fitted_dbn(dyn_gaussian_partial_fit_bytes, dyn_gaussian_fit_bytes,
                                    dyn_other_partial_fit_bytes, dyn_other_fit_bytes):
    # ####################
    # Gaussian partial fit
    # ####################
    loaded_partial = pickle.loads(dyn_gaussian_partial_fit_bytes)
    assert not loaded_partial.fitted()
    assert not loaded_partial.static_bn().fitted()
    assert not loaded_partial.transition_bn().fitted()
    cpd = loaded_partial.static_bn().cpd("d_t_1")
    assert cpd.variable() == "d_t_1"
    assert cpd.evidence() == ["a_t_2"]
    assert list(cpd.beta) == [1, 2]
    assert cpd.variance == 2

    cpd = loaded_partial.transition_bn().cpd("b_t_0")
    assert cpd.variable() == "b_t_0"
    assert cpd.evidence() == ["c_t_2"]
    assert list(cpd.beta) == [3, 4]
    assert cpd.variance == 5

    # ####################
    # Gaussian fit
    # ####################
    loaded_fitted = pickle.loads(dyn_gaussian_fit_bytes)
    assert loaded_fitted.fitted()
    assert loaded_fitted.static_bn().fitted()
    assert loaded_fitted.transition_bn().fitted()

    # ####################
    # Other partial fit
    # ####################
    loaded_partial = pickle.loads(dyn_other_partial_fit_bytes)
    assert not loaded_partial.fitted()
    assert not loaded_partial.static_bn().fitted()
    assert not loaded_partial.transition_bn().fitted()
    assert loaded_partial.static_bn().node_type("b_t_1") == pbn.DiscreteFactorType()
    assert loaded_partial.static_bn().node_type("c_t_1") == pbn.CKDEType()
    assert loaded_partial.static_bn().node_type("d_t_1") == pbn.LinearGaussianCPDType()

    assert loaded_partial.transition_bn().node_type("b_t_0") == pbn.DiscreteFactorType()
    assert loaded_partial.transition_bn().node_type("c_t_0") == pbn.CKDEType()
    assert loaded_partial.transition_bn().node_type("d_t_0") == pbn.LinearGaussianCPDType()

    cpd = loaded_partial.static_bn().cpd("d_t_1")
    assert cpd.variable() == "d_t_1"
    assert cpd.evidence() == ["a_t_2"]
    assert list(cpd.beta) == [1, 2]
    assert cpd.variance == 2

    cpd = loaded_partial.transition_bn().cpd("d_t_0")
    assert cpd.variable() == "d_t_0"
    assert cpd.evidence() == ["a_t_2"]
    assert list(cpd.beta) == [3, 4]
    assert cpd.variance == 1.5

    # ####################
    # Other fit
    # ####################
    loaded_fitted = pickle.loads(dyn_other_fit_bytes)
    assert loaded_fitted.fitted()
    assert loaded_fitted.static_bn().fitted()
    assert loaded_fitted.transition_bn().fitted()
    assert loaded_partial.static_bn().node_type("b_t_1") == pbn.DiscreteFactorType()
    assert loaded_partial.static_bn().node_type("c_t_1") == pbn.CKDEType()
    assert loaded_partial.static_bn().node_type("d_t_1") == pbn.LinearGaussianCPDType()

    assert loaded_partial.transition_bn().node_type("b_t_0") == pbn.DiscreteFactorType()
    assert loaded_partial.transition_bn().node_type("c_t_0") == pbn.CKDEType()
    assert loaded_partial.transition_bn().node_type("d_t_0") == pbn.LinearGaussianCPDType()

    cpd = loaded_partial.static_bn().cpd("d_t_1")
    assert cpd.variable() == "d_t_1"
    assert cpd.evidence() == ["a_t_2"]
    assert list(cpd.beta) == [1, 2]
    assert cpd.variance == 2

    cpd = loaded_partial.transition_bn().cpd("d_t_0")
    assert cpd.variable() == "d_t_0"
    assert cpd.evidence() == ["a_t_2"]
    assert list(cpd.beta) == [3, 4]
    assert cpd.variance == 1.5



import pytest
import pybnesian as pbn
from pybnesian import FactorType
import pickle

@pytest.fixture
def lg_type_bytes():
    lg = pbn.LinearGaussianCPDType()
    return pickle.dumps(lg)

@pytest.fixture
def ckde_type_bytes():
    ckde = pbn.CKDEType()
    return pickle.dumps(ckde)

@pytest.fixture
def discrete_type_bytes():
    discrete = pbn.DiscreteFactorType()
    return pickle.dumps(discrete)

class NewType(FactorType):
    def __init__(self):
        FactorType.__init__(self)

class OtherType(FactorType):
    def __init__(self):
        FactorType.__init__(self)

@pytest.fixture
def new_type_bytes():
    n = NewType()
    return pickle.dumps(n)

@pytest.fixture
def other_type_bytes():
    o = OtherType()
    return pickle.dumps(o)

def test_serialization_factor_type(lg_type_bytes, ckde_type_bytes, discrete_type_bytes, new_type_bytes, other_type_bytes):
    loaded_lg = pickle.loads(lg_type_bytes)
    new_lg = pbn.LinearGaussianCPDType()
    assert new_lg == loaded_lg

    loaded_ckde = pickle.loads(ckde_type_bytes)
    new_ckde = pbn.CKDEType()
    assert loaded_ckde == new_ckde

    loaded_discrete = pickle.loads(discrete_type_bytes)
    new_discrete = pbn.DiscreteFactorType()
    assert loaded_discrete == new_discrete

    loaded_new = pickle.loads(new_type_bytes)
    new_new = NewType()
    assert loaded_new == new_new

    loaded_other = pickle.loads(other_type_bytes)
    new_other = OtherType()
    assert loaded_other == new_other

    assert new_lg != new_ckde
    assert new_lg != new_discrete
    assert new_lg != new_new
    assert new_lg != new_other
    assert new_ckde != new_discrete
    assert new_ckde != new_new
    assert new_ckde != new_other
    assert new_discrete != new_new
    assert new_discrete != new_other
    assert new_new != new_other


import pytest
import numpy as np
import pyarrow as pa
import pybnesian as pbn
from pybnesian import BandwidthSelector
from scipy.stats import gaussian_kde
from functools import reduce

import util_test

SIZE = 500
df = util_test.generate_normal_data(SIZE, seed=0)
df_float = df.astype('float32')

def test_check_type():
    cpd = pbn.ProductKDE(['a'])
    cpd.fit(df)
    with pytest.raises(ValueError) as ex:
        cpd.logl(df_float)
    assert "Data type of training and test datasets is different." in str(ex.value)
    with pytest.raises(ValueError) as ex:
        cpd.slogl(df_float)
    assert "Data type of training and test datasets is different." in str(ex.value)

    cpd.fit(df_float)
    with pytest.raises(ValueError) as ex:
        cpd.logl(df)
    assert "Data type of training and test datasets is different." in str(ex.value)
    with pytest.raises(ValueError) as ex:
        cpd.slogl(df)
    assert "Data type of training and test datasets is different." in str(ex.value)

def test_productkde_variables():
    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        cpd = pbn.ProductKDE(variables)
        assert cpd.variables() == variables

def py_nr_bandwidth(df, variables):
    cov = df[variables].cov().to_numpy()
    delta = np.linalg.inv(np.diag(np.diag(cov))).dot(cov)
    delta_inv = np.linalg.inv(delta)
    N = df.shape[0]
    d = len(variables)

    k = 4*d*np.sqrt(np.linalg.det(delta))/ (2*(delta_inv.dot(delta_inv)).trace() + delta_inv.trace()**2)
    return np.power(k / N, 2 / (d + 4)) * np.diag(cov)

def py_scott_bandwidth(df, variables):
    var = df[variables].var().to_numpy()
    N = df.shape[0]
    d = len(variables)

    return np.power(N, -2 / (d + 4)) * var

def test_productkde_bandwidth():
    # for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
    for variables in [['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        for instances in [50, 150, 500]:
            cpd = pbn.ProductKDE(variables)
            cpd.fit(df.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, py_nr_bandwidth(df[:instances], variables))), "Wrong bandwidth computed with normal reference rule."

            cpd.fit(df_float.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, py_nr_bandwidth(df[:instances], variables), atol=0.0005)), "Wrong bandwidth computed with normal reference rule."

            cpd = pbn.ProductKDE(variables, pbn.ScottsBandwidth())
            cpd.fit(df.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, py_scott_bandwidth(df[:instances], variables))), "Wrong bandwidth computed with Scott's rule."

            cpd.fit(df_float.iloc[:instances])
            assert np.all(np.isclose(cpd.bandwidth, py_scott_bandwidth(df[:instances], variables), atol=0.0005)), "Wrong bandwidth computed with Scott's rule."


    cpd = pbn.ProductKDE(['a'])
    cpd.fit(df)
    cpd.bandwidth = [1]
    assert cpd.bandwidth == np.asarray([1]), "Could not change bandwidth."

    cpd.fit(df_float)
    cpd.bandwidth = [1]
    assert cpd.bandwidth == np.asarray([1]), "Could not change bandwidth."

class UnitaryBandwidth(BandwidthSelector):
    def __init__(self):
        BandwidthSelector.__init__(self)

    def diag_bandwidth(self, df, variables):
        return np.ones((len(variables),))
    
def test_productkde_new_bandwidth():
    kde = pbn.ProductKDE(["a"], UnitaryBandwidth())
    kde.fit(df)
    assert kde.bandwidth == np.ones((1,))
    
    kde.fit(df_float)
    assert kde.bandwidth == np.ones((1,))

    kde = pbn.ProductKDE(["a", "b", "c", "d"], UnitaryBandwidth())
    kde.fit(df)
    assert np.all(kde.bandwidth == np.ones((4,)))
    
    kde.fit(df_float)
    assert np.all(kde.bandwidth == np.ones((4,)))

def test_productkde_data_type():
    k = pbn.ProductKDE(["a"])

    with pytest.raises(ValueError) as ex:
        k.data_type()
    "KDE factor not fitted" in str(ex.value)

    k.fit(df)
    assert k.data_type() == pa.float64()
    k.fit(df_float)
    assert k.data_type() == pa.float32()

def test_productkde_fit():
    def _test_productkde_fit_iter(variables, _df, instances):
        cpd = pbn.ProductKDE(variables)
        assert not cpd.fitted()
        cpd.fit(_df.iloc[:instances,:])
        assert cpd.fitted()

        assert instances == cpd.num_instances(), "Wrong number of training instances."
        assert len(variables) == cpd.num_variables(), "Wrong number of training variables."
        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(cpd.bandwidth, py_nr_bandwidth(_df.iloc[:instances], variables), atol=0.0005)), "Wrong bandwidth."
        else:
            assert np.all(np.isclose(cpd.bandwidth, py_nr_bandwidth(_df.iloc[:instances], variables))), "Wrong bandwidth."

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        for instances in [50, 150, 500]:
            _test_productkde_fit_iter(variables, df, instances)
            _test_productkde_fit_iter(variables, df_float, instances)

def test_productkde_fit_null():
    def _test_productkde_fit_null_iter(variables, _df, instances):
        cpd = pbn.ProductKDE(variables)
        assert not cpd.fitted()
        cpd.fit(_df.iloc[:instances,:])
        assert cpd.fitted()

        npdata = _df.loc[:, variables].to_numpy()
        npdata_instances = npdata[:instances,:]

        nan_rows = np.any(np.isnan(npdata_instances), axis=1)
        nonnan_indices = np.where(~nan_rows)[0]

        assert (~nan_rows).sum() == cpd.num_instances(), "Wrong number of training instances with null values."
        assert len(variables) == cpd.num_variables(), "Wrong number of training variables with null values."
        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(cpd.bandwidth, py_nr_bandwidth(_df.iloc[nonnan_indices,:], variables), atol=0.0005)), "Wrong bandwidth with null values."
        else:
            assert np.all(np.isclose(cpd.bandwidth, py_nr_bandwidth(_df.iloc[nonnan_indices,:], variables))), "Wrong bandwidth with null values."

    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        for instances in [50, 150, 500]:
            _test_productkde_fit_null_iter(variables, df_null, instances)
            _test_productkde_fit_null_iter(variables, df_null_float, instances)


def factor_product_kernel(train_data):
    # Estimate bandwidth factor using Equation (3.4) of Chacon and Duong (2018)

    cov_data = np.atleast_2d(np.cov(train_data, rowvar=False, bias=False))
    delta = np.diag(np.reciprocal(np.diag(cov_data))).dot(cov_data)
    delta_inv = np.linalg.inv(delta)

    N = train_data.shape[0]
    d = train_data.shape[1]

    num_factor = 4 * d * np.sqrt(np.linalg.det(delta))
    denom_factor = (2 * np.trace(np.dot(delta_inv, delta_inv)) + np.trace(delta_inv)**2)

    k = num_factor / denom_factor

    return (k / N)**(1. / (d + 4.))


def test_productkde_logl():
    def _test_productkde_logl_iter(variables, _df, _test_df):
        cpd = pbn.ProductKDE(variables)
        cpd.fit(_df)

        logl = cpd.logl(_test_df)

        npdata = _df.loc[:, variables].to_numpy()

        factor = factor_product_kernel(npdata)

        final_scipy_kde = gaussian_kde(npdata.T, 
                                       bw_method=lambda gkde: factor * np.eye(npdata.shape[1], dtype=npdata.dtype))
        final_scipy_kde.cho_cov = np.linalg.cholesky(final_scipy_kde.covariance)
        final_scipy_kde.log_det = 2*np.log(np.diag(final_scipy_kde.cho_cov * np.sqrt(2*np.pi))).sum()

        test_npdata = _test_df.loc[:, variables].to_numpy()
        scipy = final_scipy_kde.logpdf(test_npdata.T)

        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(logl, scipy, atol=5e-3))
        else:
            assert np.all(np.isclose(logl, scipy))

    test_df = util_test.generate_normal_data(50, seed=1)
    test_df_float = test_df.astype('float32')

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_productkde_logl_iter(variables, df, test_df)
        _test_productkde_logl_iter(variables, df_float, test_df_float)

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.logl(test_df), cpd2.logl(test_df))), "Order of evidence changes logl() result."

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.logl(test_df_float), cpd2.logl(test_df_float), atol=0.0005)), "Order of evidence changes logl() result."

def test_productkde_logl_null():
    def _test_productkde_logl_null_iter(variables, _df, _test_df):
        cpd = pbn.ProductKDE(variables)
        cpd.fit(_df)

        logl = cpd.logl(_test_df)

        npdata = _df.loc[:, variables].to_numpy()

        factor = factor_product_kernel(npdata)
        final_scipy_kde = gaussian_kde(npdata.T, 
                                       bw_method=lambda gkde: factor * np.eye(npdata.shape[1], dtype=npdata.dtype))
        final_scipy_kde.cho_cov = np.linalg.cholesky(final_scipy_kde.covariance)
        final_scipy_kde.log_det = 2*np.log(np.diag(final_scipy_kde.cho_cov * np.sqrt(2*np.pi))).sum()

        test_npdata = _test_df.loc[:, variables].to_numpy()

        scipy = np.full((test_npdata.shape[0],), np.nan)
        nan_rows = np.any(np.isnan(test_npdata), axis=1)
        scipy[~nan_rows] = final_scipy_kde.logpdf(test_npdata[~nan_rows].T)

        if npdata.dtype == "float32":
            assert np.all(np.isclose(logl, scipy, atol=5e-3, equal_nan=True))
        else:
            assert np.all(np.isclose(logl, scipy, equal_nan=True))

    TEST_SIZE = 50

    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_productkde_logl_null_iter(variables, df, df_null)
        _test_productkde_logl_null_iter(variables, df_float, df_null_float)

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.logl(df_null), cpd2.logl(df_null), equal_nan=True)), "Order of evidence changes logl() result."

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.logl(df_null_float), cpd2.logl(df_null_float), atol=0.0005, equal_nan=True)), "Order of evidence changes logl() result."

def test_productkde_slogl():
    def _test_productkde_slogl_iter(variables, _df, _test_df):
        cpd = pbn.ProductKDE(variables)
        cpd.fit(_df)

        npdata = _df.loc[:, variables].to_numpy()
        
        factor = factor_product_kernel(npdata)
        final_scipy_kde = gaussian_kde(npdata.T, 
                                       bw_method=lambda gkde: factor * np.eye(npdata.shape[1], dtype=npdata.dtype))
        final_scipy_kde.cho_cov = np.linalg.cholesky(final_scipy_kde.covariance)
        final_scipy_kde.log_det = 2*np.log(np.diag(final_scipy_kde.cho_cov * np.sqrt(2*np.pi))).sum()

        test_npdata = _test_df.loc[:, variables].to_numpy()

        if np.all(_df.dtypes == 'float32'):
            assert np.all(np.isclose(cpd.slogl(_test_df), final_scipy_kde.logpdf(test_npdata.T).sum(),
                                     atol=5e-3*test_npdata.shape[0]))
        else:
            assert np.all(np.isclose(cpd.slogl(_test_df), final_scipy_kde.logpdf(test_npdata.T).sum()))

    test_df = util_test.generate_normal_data(50, seed=1)
    test_df_float = test_df.astype('float32')

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_productkde_slogl_iter(variables, df, test_df)
        _test_productkde_slogl_iter(variables, df_float, test_df_float)

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.slogl(test_df), cpd2.slogl(test_df))), "Order of evidence changes slogl() result."

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.slogl(test_df_float), cpd2.slogl(test_df_float), atol=0.0005)), "Order of evidence changes slogl() result."


def test_productkde_slogl_null():
    def _test_productkde_slogl_null_iter(variables, _df, _test_df):
        cpd = pbn.ProductKDE(variables)
        cpd.fit(_df)

        npdata = _df.loc[:, variables].to_numpy()
        
        factor = factor_product_kernel(npdata)
        final_scipy_kde = gaussian_kde(npdata.T, 
                                       bw_method=lambda gkde: factor * np.eye(npdata.shape[1], dtype=npdata.dtype))
        final_scipy_kde.cho_cov = np.linalg.cholesky(final_scipy_kde.covariance)
        final_scipy_kde.log_det = 2*np.log(np.diag(final_scipy_kde.cho_cov * np.sqrt(2*np.pi))).sum()

        test_npdata = _test_df.loc[:, variables].to_numpy()
        nan_rows = np.any(np.isnan(test_npdata), axis=1)

        if npdata.dtype == "float32":
            assert np.all(np.isclose(cpd.slogl(_test_df),
                                     np.sum(final_scipy_kde.logpdf(test_npdata[~nan_rows].T)),
                                     atol=5e-3*test_npdata.shape[0]))
        else:
            assert np.all(np.isclose(cpd.slogl(_test_df),
                                     np.sum(final_scipy_kde.logpdf(test_npdata[~nan_rows].T))))

    TEST_SIZE = 50

    test_df = util_test.generate_normal_data(TEST_SIZE, seed=1)
    test_df_float = test_df.astype('float32')

    np.random.seed(0)
    a_null = np.random.randint(0, TEST_SIZE, size=10)
    b_null = np.random.randint(0, TEST_SIZE, size=10)
    c_null = np.random.randint(0, TEST_SIZE, size=10)
    d_null = np.random.randint(0, TEST_SIZE, size=10)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    df_null_float = test_df_float.copy()
    df_null_float.loc[df_null_float.index[a_null], 'a'] = np.nan
    df_null_float.loc[df_null_float.index[b_null], 'b'] = np.nan
    df_null_float.loc[df_null_float.index[c_null], 'c'] = np.nan
    df_null_float.loc[df_null_float.index[d_null], 'd'] = np.nan

    for variables in [['a'], ['b', 'a'], ['c', 'a', 'b'], ['d', 'a', 'b', 'c']]:
        _test_productkde_slogl_null_iter(variables, df, df_null)
        _test_productkde_slogl_null_iter(variables, df_float, df_null_float)


    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df)
    assert np.all(np.isclose(cpd.slogl(df_null), cpd2.slogl(df_null))), "Order of evidence changes slogl() result."

    cpd = pbn.ProductKDE(['d', 'a', 'b', 'c'])
    cpd.fit(df_float)
    cpd2 = pbn.ProductKDE(['a', 'c', 'd', 'b'])
    cpd2.fit(df_float)
    assert np.all(np.isclose(cpd.slogl(df_null_float), cpd2.slogl(df_null_float), atol=0.0005)), "Order of evidence changes slogl() result."



import numpy as np
import pandas as pd


def generate_normal_data(size, seed=0):
    np.random.seed(seed)

    a_array = np.random.normal(3, 0.5, size=size)
    b_array = 2.5 + 1.65*a_array + np.random.normal(0, 2, size=size)
    c_array = -4.2 - 1.2*a_array + 3.2*b_array + np.random.normal(0, 0.75, size=size)
    d_array = 1.5 - 0.9*a_array + 5.6*b_array + 0.3 * c_array + np.random.normal(0, 0.5, size=size)


    return pd.DataFrame({
                    'a': a_array,
                    'b': b_array,
                    'c': c_array,
                    'd': d_array
                    })


def generate_normal_data_indep(size, seed=0):
    np.random.seed(seed)

    a_array = np.random.normal(3, 0.5, size=size)
    b_array = np.random.normal(2.5, 2, size=size)
    c_array = -4.2 - 1.2*a_array + 3.2*b_array + np.random.normal(0, 0.75, size=size)
    d_array = 1.5 - 0.3 * c_array + np.random.normal(0, 0.5, size=size)


    return pd.DataFrame({
                    'a': a_array,
                    'b': b_array,
                    'c': c_array,
                    'd': d_array
                    })


def generate_discrete_data_uniform(size, seed=0):
    np.random.seed(seed)

    a_dict = np.asarray(['a1', 'a2'])
    b_dict = np.asarray(['b1', 'b2', 'b3'])
    c_dict = np.asarray(['c1', 'c2'])
    d_dict = np.asarray(['d1', 'd2', 'd3', 'd4'])

    return pd.DataFrame({'A': a_dict[np.random.randint(0, a_dict.size, size=size)],
                         'B': b_dict[np.random.randint(0, b_dict.size, size=size)],
                         'C': c_dict[np.random.randint(0, c_dict.size, size=size)],
                         'D': d_dict[np.random.randint(0, d_dict.size, size=size)]
                        }, dtype='category')


def generate_discrete_data_dependent(size, seed=0):
    np.random.seed(seed)

    a_dict = np.asarray(['a1', 'a2'])
    b_dict = np.asarray(['b1', 'b2', 'b3'])
    c_dict = np.asarray(['c1', 'c2'])
    d_dict = np.asarray(['d1', 'd2', 'd3', 'd4'])

    a_values = a_dict[np.random.choice(a_dict.size, size, p=[0.75, 0.25])]
    b_values = np.empty_like(a_values)
    c_values = np.empty_like(a_values)
    d_values = np.empty_like(a_values)

    a1_indices = a_values == 'a1'

    b_values[a1_indices] = b_dict[np.random.choice(b_dict.size, np.sum(a1_indices), p=[0.33, 0.33, 0.34])]
    b_values[~a1_indices] = b_dict[np.random.choice(b_dict.size, np.sum(~a1_indices), p=[0, 0.8, 0.2])]

    a1b1_indices = np.logical_and(a_values == 'a1', b_values == 'b1')
    a1b2_indices = np.logical_and(a_values == 'a1', b_values == 'b2')
    a1b3_indices = np.logical_and(a_values == 'a1', b_values == 'b3')
    a2b1_indices = np.logical_and(a_values == 'a2', b_values == 'b1')
    a2b2_indices = np.logical_and(a_values == 'a2', b_values == 'b2')
    a2b3_indices = np.logical_and(a_values == 'a2', b_values == 'b3')

    c_values[a1b1_indices] = c_dict[np.random.choice(c_dict.size, np.sum(a1b1_indices), p=[0.5, 0.5])]
    c_values[a1b2_indices] = c_dict[np.random.choice(c_dict.size, np.sum(a1b2_indices), p=[0.75, 0.25])]
    c_values[a1b3_indices] = c_dict[np.random.choice(c_dict.size, np.sum(a1b3_indices), p=[0.2, 0.8])]
    c_values[a2b1_indices] = c_dict[np.random.choice(c_dict.size, np.sum(a2b1_indices), p=[1, 0])]
    c_values[a2b2_indices] = c_dict[np.random.choice(c_dict.size, np.sum(a2b2_indices), p=[0, 1])]
    c_values[a2b3_indices] = c_dict[np.random.choice(c_dict.size, np.sum(a2b3_indices), p=[0.01, 0.99])]

    c1_indices = c_values == 'c1'
    c2_indices = c_values == 'c2'

    d_values[c1_indices] = d_dict[np.random.choice(d_dict.size, np.sum(c1_indices), p=[0.25, 0.25, 0.25, 0.25])]
    d_values[c2_indices] = d_dict[np.random.choice(d_dict.size, np.sum(c2_indices), p=[0.7, 0, 0.15, 0.15])]

    return pd.DataFrame({'A': a_values,
                         'B': b_values,
                         'C': c_values,
                         'D': d_values
                        }, dtype='category')


def generate_hybrid_data(size, seed=0):
    #
    #   Generate data from:
    #
    #   A   B   C
    #    \  |  /
    #     \ | /
    #       v
    #       D
    np.random.seed(seed)

    a_dict = np.asarray(['a1', 'a2'])
    a_values = a_dict[np.random.choice(a_dict.size, size, p=[0.75, 0.25])]

    b_dict = np.asarray(['b1', 'b2', 'b3'])
    b_values = b_dict[np.random.choice(b_dict.size, size, p=[0.3, 0.4, 0.3])]

    c_values = -4.2 + np.random.normal(0, 0.75, size=size)

    a1b1_indices = np.logical_and(a_values == 'a1', b_values == 'b1')
    a1b2_indices = np.logical_and(a_values == 'a1', b_values == 'b2')
    a1b3_indices = np.logical_and(a_values == 'a1', b_values == 'b3')
    a2b1_indices = np.logical_and(a_values == 'a2', b_values == 'b1')
    a2b2_indices = np.logical_and(a_values == 'a2', b_values == 'b2')
    a2b3_indices = np.logical_and(a_values == 'a2', b_values == 'b3')

    d_values = np.empty_like(c_values)
    d_values[a1b1_indices] = np.random.normal(1, 0.75, size=a1b1_indices.sum())
    d_values[a1b2_indices] = -2 + c_values[a1b2_indices] + np.random.normal(0, 2, size=a1b2_indices.sum())
    d_values[a1b3_indices] = -1 + 3*c_values[a1b3_indices] + np.random.normal(0, 0.25, size=a1b3_indices.sum())
    d_values[a2b1_indices] = np.random.normal(2, 1, size=a2b1_indices.sum())
    d_values[a2b2_indices] = 3.5 + -1.2*c_values[a2b2_indices] + np.random.normal(0, 1, size=a2b2_indices.sum())
    d_values[a2b3_indices] = 4.8 + -2*c_values[a2b3_indices] + np.random.normal(0, 1.5, size=a2b3_indices.sum())

    return pd.DataFrame({'A': pd.Series(a_values, dtype='category'),
                         'B': pd.Series(b_values, dtype='category'),
                         'C': c_values,
                         'D': d_values
                        })

def generate_indep_hybrid_data(size, seed=0):
    np.random.seed(seed)

    d2_dict = np.asarray(['a1', 'a2'])
    d2_values = d2_dict[np.random.choice(d2_dict.size, size, p=[0.5, 0.5])]

    d3_dict = np.asarray(['b1', 'b2', 'b3'])
    d3_values = d3_dict[np.random.choice(d3_dict.size, size, p=[0.33, 0.34, 0.33])]

    d4_dict = np.asarray(['c1', 'c2', 'c3', 'c4'])
    d4_values = d4_dict[np.random.choice(d4_dict.size, size, p=[0.25, 0.25, 0.25, 0.25])]

    d5_dict = np.asarray(['d1', 'd2', 'd3', 'd4', 'd5'])
    d5_values = d5_dict[np.random.choice(d5_dict.size, size, p=[0.2, 0.2, 0.2, 0.2, 0.2])]

    d6_dict = np.asarray(['e1', 'e2', 'e3', 'e4', 'e5', 'e6'])
    d6_values = d6_dict[np.random.choice(d6_dict.size, size, p=[0.166, 0.166, 0.166, 0.166, 0.166, 0.17])]

    c1_values = -4.2 + np.random.normal(0, 0.75, size=size)
    c2_values = np.random.normal(1, 2, size=size)
    c3_values = np.random.normal(2, 0.7, size=size)
    c4_values = np.random.normal(-3, 2.5, size=size)
    c5_values = np.random.normal(-1.2, 0.5, size=size)
    c6_values = np.random.normal(3, 1.5, size=size)

    return pd.DataFrame({'D2': pd.Series(d2_values, dtype='category'),
                         'D3': pd.Series(d3_values, dtype='category'),
                         'D4': pd.Series(d4_values, dtype='category'),
                         'D5': pd.Series(d5_values, dtype='category'),
                         'D6': pd.Series(d6_values, dtype='category'),
                         'C1': c1_values,
                         'C2': c2_values,
                         'C3': c3_values,
                         'C4': c4_values,
                         'C5': c5_values,
                         'C6': c6_values,
                        })


import pytest
import numpy as np
import pybnesian as pbn
from pybnesian import BayesianNetwork, GaussianNetwork
import util_test

df = util_test.generate_normal_data(10000)

def test_create_bn():
    gbn = GaussianNetwork(['a', 'b', 'c', 'd'])

    assert gbn.num_nodes() == 4
    assert gbn.num_arcs() == 0
    assert gbn.nodes() == ['a', 'b', 'c', 'd']

    gbn = GaussianNetwork(['a', 'b', 'c', 'd'], [('a', 'c')])
    assert gbn.num_nodes() == 4
    assert gbn.num_arcs() == 1
    assert gbn.nodes() == ['a', 'b', 'c', 'd']

    gbn = GaussianNetwork([('a', 'c'), ('b', 'd'), ('c', 'd')])
    assert gbn.num_nodes() == 4
    assert gbn.num_arcs() == 3
    assert gbn.nodes() == ['a', 'c', 'b', 'd']

    with pytest.raises(TypeError) as ex:
        gbn = GaussianNetwork(['a', 'b', 'c'], [('a', 'c', 'b')])
    assert "incompatible constructor arguments" in str(ex.value)
    
    with pytest.raises(IndexError) as ex:
        gbn = GaussianNetwork(['a', 'b', 'c'], [('a', 'd')])
    assert "not present in the graph" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn = GaussianNetwork([('a', 'b'), ('b', 'c'), ('c', 'a')])
    assert "must be a DAG" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn = GaussianNetwork(['a', 'b', 'c', 'd'], [('a', 'b'), ('b', 'c'), ('c', 'a')])
    assert "must be a DAG" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn = BayesianNetwork(pbn.GaussianNetworkType(), ['a', 'b', 'c', 'd'], [], [('a', pbn.CKDEType())])
    assert "Wrong factor type" in str(ex.value)
    
def gbn_generator():
    # Test different Networks created with different constructors.
    gbn = GaussianNetwork(['a', 'b', 'c', 'd'])
    yield gbn
    gbn = GaussianNetwork([('a', 'c'), ('b', 'd'), ('c', 'd')])
    yield gbn
    gbn = GaussianNetwork(['a', 'b', 'c', 'd'], [('a', 'b'), ('b', 'c')])
    yield gbn

def test_nodes_util():
    for gbn in gbn_generator():
        assert gbn.num_nodes() == 4

        nodes = gbn.nodes()
        indices = gbn.indices()

        assert nodes[gbn.index('a')] == 'a'
        assert nodes[gbn.index('b')] == 'b'
        assert nodes[gbn.index('c')] == 'c'
        assert nodes[gbn.index('d')] == 'd'

        assert indices[gbn.name(0)] == 0
        assert indices[gbn.name(1)] == 1
        assert indices[gbn.name(2)] == 2
        assert indices[gbn.name(3)] == 3

        assert gbn.contains_node('a')
        assert gbn.contains_node('b')
        assert gbn.contains_node('c')
        assert gbn.contains_node('d')
        assert not gbn.contains_node('e')

def test_parent_children():
    gbn = GaussianNetwork(['a', 'b', 'c', 'd'])
    
    assert gbn.num_parents('a') == 0
    assert gbn.num_parents('b') == 0
    assert gbn.num_parents('c') == 0
    assert gbn.num_parents('d') == 0

    assert gbn.parents('a') == []
    assert gbn.parents('b') == []
    assert gbn.parents('c') == []
    assert gbn.parents('d') == []

    assert gbn.num_children('a') == 0
    assert gbn.num_children('b') == 0
    assert gbn.num_children('c') == 0
    assert gbn.num_children('d') == 0

    gbn = GaussianNetwork([('a', 'c'), ('b', 'd'), ('c', 'd')])

    assert gbn.num_parents('a') == 0
    assert gbn.num_parents('b') == 0
    assert gbn.num_parents('c') == 1
    assert gbn.num_parents('d') == 2

    assert gbn.parents('a') == []
    assert gbn.parents('b') == []
    assert gbn.parents('c') == ['a']
    assert set(gbn.parents('d')) == set(['b', 'c'])

    assert gbn.num_children('a') == 1
    assert gbn.num_children('b') == 1
    assert gbn.num_children('c') == 1
    assert gbn.num_children('d') == 0

    gbn = GaussianNetwork(['a', 'b', 'c', 'd'], [('a', 'b'), ('b', 'c')])

    assert gbn.num_parents('a') == 0
    assert gbn.num_parents('b') == 1
    assert gbn.num_parents('c') == 1
    assert gbn.num_parents('d') == 0

    assert gbn.parents('a') == []
    assert gbn.parents('b') == ['a']
    assert gbn.parents('c') == ['b']
    assert gbn.parents('d') == []

    assert gbn.num_children('a') == 1
    assert gbn.num_children('b') == 1
    assert gbn.num_children('c') == 0
    assert gbn.num_children('d') == 0

def test_arcs():
    gbn = GaussianNetwork(['a', 'b', 'c', 'd'])

    assert gbn.num_arcs() == 0
    assert gbn.arcs() == []
    assert not gbn.has_arc('a', 'b')

    gbn.add_arc('a', 'b')
    assert gbn.num_arcs() == 1
    assert gbn.arcs() == [('a', 'b')]
    assert gbn.parents('b') == ['a']
    assert gbn.num_parents('b') == 1
    assert gbn.num_children('a') == 1
    assert gbn.has_arc('a', 'b')

    gbn.add_arc('b', 'c')
    assert gbn.num_arcs() == 2
    assert set(gbn.arcs()) == set([('a', 'b'), ('b', 'c')])
    assert gbn.parents('c') == ['b']
    assert gbn.num_parents('c') == 1
    assert gbn.num_children('b') == 1
    assert gbn.has_arc('b', 'c')
    
    gbn.add_arc('d', 'c')
    assert gbn.num_arcs() == 3
    assert set(gbn.arcs()) == set([('a', 'b'), ('b', 'c'), ('d', 'c')])
    assert set(gbn.parents('c')) == set(['b', 'd'])
    assert gbn.num_parents('c') == 2
    assert gbn.num_children('d') == 1
    assert gbn.has_arc('d', 'c')

    assert gbn.has_path('a', 'c')
    assert not gbn.has_path('a', 'd')
    assert gbn.has_path('b', 'c')
    assert gbn.has_path('d', 'c')

    assert not gbn.can_add_arc('c', 'a')
    # This edge exists, but virtually we consider that the addition is allowed. 
    assert gbn.can_add_arc('b', 'c')
    assert gbn.can_add_arc('d', 'a')

    gbn.add_arc('b', 'd')
    assert gbn.num_arcs() == 4
    assert set(gbn.arcs()) == set([('a', 'b'), ('b', 'c'), ('d', 'c'), ('b', 'd')])
    assert gbn.parents('d') == ['b']
    assert gbn.num_parents('d') == 1
    assert gbn.num_children('b') == 2
    assert gbn.has_arc('b', 'd')

    assert gbn.has_path('a', 'd')
    assert not gbn.can_add_arc('d', 'a')
    assert not gbn.can_flip_arc('b', 'c')
    assert gbn.can_flip_arc('a', 'b')
    # This edge does not exist, but it could be flipped if it did.
    assert gbn.can_flip_arc('d', 'a')

    # We can add an edge twice without changes.
    gbn.add_arc('b', 'd')
    assert gbn.num_arcs() == 4
    assert set(gbn.arcs()) == set([('a', 'b'), ('b', 'c'), ('d', 'c'), ('b', 'd')])
    assert gbn.parents('d') == ['b']
    assert gbn.num_parents('d') == 1
    assert gbn.num_children('b') == 2
    assert gbn.has_arc('b', 'd')

    gbn.remove_arc('b', 'c')
    assert gbn.num_arcs() == 3
    assert set(gbn.arcs()) == set([('a', 'b'), ('d', 'c'), ('b', 'd')])
    assert gbn.parents('c') == ['d']
    assert gbn.num_parents('c') == 1
    assert gbn.num_children('b') == 1
    assert not gbn.has_arc('b', 'c')

    assert gbn.can_add_arc('b', 'c')
    assert not gbn.can_add_arc('c', 'b')
    assert gbn.has_path('a', 'c')
    assert gbn.has_path('b', 'c')

    gbn.remove_arc('d', 'c')    
    assert gbn.num_arcs() == 2
    assert set(gbn.arcs()) == set([('a', 'b'), ('b', 'd')])
    assert gbn.parents('c') == []
    assert gbn.num_parents('c') == 0
    assert gbn.num_children('d') == 0
    assert not gbn.has_arc('d', 'c')

    assert gbn.can_add_arc('b', 'c')
    assert gbn.can_add_arc('c', 'b')
    assert not gbn.has_path('a', 'c')
    assert not gbn.has_path('b', 'c')

def test_bn_fit():
    gbn = GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    with pytest.raises(ValueError) as ex:
        for n in gbn.nodes():
            cpd = gbn.cpd(n)
    assert "not added" in str(ex.value)

    gbn.fit(df)

    for n in gbn.nodes():
        cpd = gbn.cpd(n)
        assert cpd.variable() == n
        assert cpd.evidence() == gbn.parents(n)

    gbn.fit(df)
    
    gbn.remove_arc('a', 'b')

    cpd_b = gbn.cpd('b')
    assert cpd_b.evidence != gbn.parents('b')

    gbn.fit(df)

    cpd_b = gbn.cpd('b')
    assert cpd_b.evidence() == gbn.parents('b')

def test_add_cpds():
    gbn = GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    with pytest.raises(ValueError) as ex:
        gbn.add_cpds([pbn.LinearGaussianCPD('e', [])])
    assert "variable which is not present" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn.add_cpds([pbn.LinearGaussianCPD('a', ['e'])])
    assert "Evidence variable" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn.add_cpds([pbn.LinearGaussianCPD('a', ['b'])])
    assert "CPD do not have the model's parent set as evidence" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn.add_cpds([pbn.LinearGaussianCPD('b', [])])
    assert "CPD do not have the model's parent set as evidence" in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn.add_cpds([pbn.LinearGaussianCPD('b', ['c'])])
    assert "CPD do not have the model's parent set as evidence" in str(ex.value)

    lg = pbn.LinearGaussianCPD('b', ['a'], [2.5, 1.65], 4)
    assert lg.fitted()

    gbn.add_cpds([lg])

    cpd_b = gbn.cpd('b')
    assert cpd_b.variable() == 'b'
    assert cpd_b.evidence() == ['a']
    assert cpd_b.fitted()
    assert np.all(cpd_b.beta == np.asarray([2.5, 1.65]))
    assert cpd_b.variance == 4

    with pytest.raises(ValueError) as ex:
        cpd_a = gbn.cpd('a')
    assert "CPD of variable \"a\" not added. Call add_cpds() or fit() to add the CPD." in str(ex.value)

    with pytest.raises(ValueError) as ex:
        cpd_c = gbn.cpd('c')
    assert "CPD of variable \"c\" not added. Call add_cpds() or fit() to add the CPD." in str(ex.value)

    with pytest.raises(ValueError) as ex:
        cpd_d = gbn.cpd('d')
    assert "CPD of variable \"d\" not added. Call add_cpds() or fit() to add the CPD." in str(ex.value)

    with pytest.raises(ValueError) as ex:
        gbn.add_cpds([pbn.LinearGaussianCPD('e', [])])
    assert "variable which is not present" in str(ex.value)


def test_bn_logl():
    gbn = GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    gbn.fit(df)

    test_df = util_test.generate_normal_data(5000)
    ll = gbn.logl(test_df)
    sll = gbn.slogl(test_df)

    sum_ll = np.zeros((5000,))
    sum_sll = 0
    
    for n in gbn.nodes():
        cpd = gbn.cpd(n)
        l = cpd.logl(test_df)
        s = cpd.slogl(test_df)
        assert np.all(np.isclose(s, l.sum()))
        sum_ll += l
        sum_sll += s
    
    assert np.all(np.isclose(ll, sum_ll))
    assert np.isclose(sll, ll.sum())
    assert sll == sum_sll

def test_bn_sample():
    gbn = GaussianNetwork(['a', 'c', 'b', 'd'], [('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    gbn.fit(df)
    sample = gbn.sample(1000, 0, False)

    # Not ordered, so topological sort.
    assert sample.schema.names == ['a', 'b', 'c', 'd']
    assert sample.num_rows == 1000
    
    sample_ordered = gbn.sample(1000, 0, True)
    assert sample_ordered.schema.names == ['a', 'c', 'b', 'd']
    assert sample_ordered.num_rows == 1000

    assert sample.column(0).equals(sample_ordered.column(0))
    assert sample.column(1).equals(sample_ordered.column(2))
    assert sample.column(2).equals(sample_ordered.column(1))
    assert sample.column(3).equals(sample_ordered.column(3))

    other_seed = gbn.sample(1000, 1, False)

    assert not sample.column(0).equals(other_seed.column(0))
    assert not sample.column(1).equals(other_seed.column(2))
    assert not sample.column(2).equals(other_seed.column(1))
    assert not sample.column(3).equals(other_seed.column(3))



import pytest
import pybnesian as pbn

def test_create():
    o = pbn.AddArc("a", "b", 1)
    assert o.source() == 'a'
    assert o.target() == 'b'
    assert o.delta() == 1

    o = pbn.RemoveArc("a", "b", 2)
    assert o.source() == 'a'
    assert o.target() == 'b'
    assert o.delta() == 2

    o = pbn.FlipArc("a", "b", 3)
    assert o.source() == 'a'
    assert o.target() == 'b'
    assert o.delta() == 3

    o = pbn.ChangeNodeType("a", pbn.CKDEType(), 4)
    assert o.node() == 'a'
    assert o.node_type() == pbn.CKDEType()
    assert o.delta() == 4

def test_apply():
    gbn = pbn.GaussianNetwork(['a', 'b', 'c', 'd'])
    assert gbn.num_arcs() == 0
    assert not gbn.has_arc('a', 'b')

    o = pbn.AddArc("a", "b", 1)
    o.apply(gbn)
    assert gbn.num_arcs() == 1
    assert gbn.has_arc('a', 'b')
    
    o = pbn.FlipArc("a", "b", 1)
    o.apply(gbn)
    assert gbn.num_arcs() == 1
    assert not gbn.has_arc('a', 'b')
    assert gbn.has_arc('b', 'a')

    o = pbn.RemoveArc("b", "a", 1)
    o.apply(gbn)
    assert gbn.num_arcs() == 0
    assert not gbn.has_arc('b', 'a')

    o = pbn.ChangeNodeType("a", pbn.CKDEType(), 1)
    with pytest.raises(ValueError) as ex:
        o.apply(gbn)
    assert "Wrong factor type" in str(ex.value)

    spbn = pbn.SemiparametricBN(['a', 'b', 'c', 'd'])
    assert spbn.num_arcs() == 0

    o = pbn.ChangeNodeType("a", pbn.CKDEType(), 1)
    assert(spbn.node_type('a') == pbn.UnknownFactorType())
    o.apply(spbn)
    assert(spbn.node_type('a') == pbn.CKDEType())

    assert not spbn.has_arc('a', 'b')
    o = pbn.AddArc("a", "b", 1)
    o.apply(spbn)
    assert spbn.num_arcs() == 1
    assert spbn.has_arc('a', 'b')
    
    o = pbn.FlipArc("a", "b", 1)
    o.apply(spbn)
    assert spbn.num_arcs() == 1
    assert not spbn.has_arc('a', 'b')
    assert spbn.has_arc('b', 'a')

    o = pbn.RemoveArc("b", "a", 1)
    o.apply(spbn)
    assert spbn.num_arcs() == 0
    assert not spbn.has_arc('b', 'a')

def test_opposite():
    bn = pbn.SemiparametricBN(["a", "b"])
    o = pbn.AddArc("a", "b", 1)
    oppo = o.opposite(bn)
    assert oppo.source() == 'a'
    assert oppo.target() == 'b'
    assert oppo.delta() == -1
    assert type(oppo) == pbn.RemoveArc

    o = pbn.RemoveArc("a", "b", 1)
    oppo = o.opposite(bn)
    assert oppo.source() == 'a'
    assert oppo.target() == 'b'
    assert oppo.delta() == -1
    assert type(oppo) == pbn.AddArc

    o = pbn.FlipArc("a", "b", 1)
    oppo = o.opposite(bn)
    assert oppo.source() == 'b'
    assert oppo.target() == 'a'
    assert oppo.delta() == -1
    assert type(oppo) == pbn.FlipArc

    bn.set_node_type("a", pbn.LinearGaussianCPDType())
    o = pbn.ChangeNodeType("a", pbn.CKDEType(), 1)
    oppo = o.opposite(bn)
    assert oppo.node() == 'a'
    assert oppo.node_type() == pbn.LinearGaussianCPDType()
    assert oppo.delta() == -1
    assert type(oppo) == pbn.ChangeNodeType


import pytest
import numpy as np
import pybnesian as pbn
import util_test

SIZE = 10000
df = util_test.generate_normal_data(SIZE)

def test_create_change_node():
    gbn = pbn.GaussianNetwork(['a', 'b', 'c', 'd'])
    
    cv = pbn.CVLikelihood(df)

    node_op = pbn.ChangeNodeTypeSet()

    with pytest.raises(ValueError) as ex:
        node_op.cache_scores(gbn, cv)
    assert "can only be used with non-homogeneous" in str(ex.value)

def test_lists():
    gbn = pbn.GaussianNetwork(['a', 'b', 'c', 'd'])
    bic = pbn.BIC(df)
    arc_op = pbn.ArcOperatorSet()

    arc_op.set_arc_blacklist([("b", "a")])
    arc_op.set_arc_whitelist([("b", "c")])
    arc_op.set_max_indegree(3)
    arc_op.set_type_whitelist([("a", pbn.LinearGaussianCPDType())])

    arc_op.cache_scores(gbn, bic)

    arc_op.set_arc_blacklist([("e", "a")])

    with pytest.raises(ValueError) as ex:
        arc_op.cache_scores(gbn, bic)
    assert "not present in the graph" in str(ex.value)

    arc_op.set_arc_whitelist([("e", "a")])

    with pytest.raises(ValueError) as ex:
        arc_op.cache_scores(gbn, bic)
    assert "not present in the graph" in str(ex.value)


def test_check_max_score():
    gbn = pbn.GaussianNetwork(['c', 'd'])

    bic = pbn.BIC(df)
    arc_op = pbn.ArcOperatorSet()

    arc_op.cache_scores(gbn, bic)
    op = arc_op.find_max(gbn)

    assert np.isclose(op.delta(), (bic.local_score(gbn, 'd', ['c']) - bic.local_score(gbn, 'd')))

    # BIC is decomposable so the best operation is the arc in reverse direction.
    arc_op.set_arc_blacklist([(op.source(), op.target())])
    arc_op.cache_scores(gbn, bic)
    
    op2 = arc_op.find_max(gbn)

    assert op.source() == op2.target()
    assert op.target() == op2.source()
    assert (type(op) == type(op2)) and (type(op) == pbn.AddArc)

def test_nomax():
    gbn = pbn.GaussianNetwork(['a', 'b'])

    bic = pbn.BIC(df)
    arc_op = pbn.ArcOperatorSet(whitelist=[("a", "b")])
    arc_op.cache_scores(gbn, bic)

    op = arc_op.find_max(gbn)

    assert op is None






import numpy as np
import pybnesian as pbn

import util_test

SIZE = 10000

df = util_test.generate_normal_data(SIZE)

def test_cv_disjoint_indices():
    cv = pbn.CrossValidation(df)

    for (train_df, test_df), (train_indices, test_indices) in zip(cv, cv.indices()):
        nptrain = np.asarray(train_indices)
        nptest = np.asarray(test_indices)
        combination = np.hstack((nptrain, nptest))
        
        assert np.all(np.sort(combination) == np.arange(SIZE)), "Not all the examples are included in the cross validation."
        assert np.all(train_df.to_pandas().to_numpy() == df.iloc[train_indices,:].to_numpy()), \
                                "The CV iterator do not slice the train dataset exactly equal as the CV indices iterator."
        assert np.all(test_df.to_pandas().to_numpy() == df.iloc[test_indices,:].to_numpy()), \
                                "The CV iterator do not slice the test dataset exactly equal as the CV indices iterator."

        assert np.setdiff1d(nptrain, nptest).shape == nptrain.shape, "The train indices includes test indices"
        assert np.setdiff1d(nptest, nptrain).shape == nptest.shape, "The test indices includes train indices"
        assert np.all(np.sort(np.setdiff1d(train_indices, test_indices)) == np.sort(train_indices)), "The train indices includes test indices"
        assert np.all(np.sort(np.setdiff1d(test_indices, train_indices)) == np.sort(test_indices)), "The test indices includes train indices"


def test_cv_fold():
    cv = pbn.CrossValidation(df)

    for i, (train_df, test_df) in enumerate(cv):
        train_fold, test_fold = cv.fold(i)

        assert train_fold.equals(train_df), "Train DataFrame fold() and __iter__ are not equal."
        assert test_fold.equals(test_df), "Test DataFrame fold() and __iter__ are not equal."


def test_cv_seed():
    cv = pbn.CrossValidation(df, seed=0)
    
    dataframes = list(cv)

    cv2 = pbn.CrossValidation(df, seed=0)

    for (train_cv, test_cv), (train_cv2, test_cv2) in zip(dataframes, cv2):
        assert train_cv.equals(train_cv2), "Train CV DataFrames with the same seed are not equal."
        assert test_cv.equals(test_cv2), "Test CV DataFrames with the same seed are not equal."

    cv3 = pbn.CrossValidation(df, seed=1)
    for (train_cv2, test_cv2), (train_cv3, test_cv3) in zip(cv2, cv3):
        assert not train_cv2.equals(train_cv3), "Train CV DataFrames with different seeds return the same result."
        assert not test_cv2.equals(test_cv3), "Test CV DataFrames with different seeds return the same result."

def test_cv_num_folds():
    cv = pbn.CrossValidation(df)
    
    dataframes = list(cv)
    indices = list(cv.indices())

    assert len(dataframes) == 10, "Default number of folds must be 10."
    assert len(indices) == 10, "Default number of folds must be 10."
    
    cv5 = pbn.CrossValidation(df, 5)
    dataframes = list(cv5)
    indices = list(cv5.indices())
    assert len(dataframes) == 5, "Wrong number of folds"
    assert len(indices) == 5, "Wrong number of folds for the indices iterator."


def test_cv_loc():
    cv = pbn.CrossValidation(df)
    
    for (train_df, test_df) in cv.loc("a"):
        assert train_df.num_columns == 1, "Only column \"a\" must be present in train DataFrame."
        assert test_df.num_columns == 1, "Only column \"a\" must be present in test DataFrame."
        train_schema = train_df.schema
        test_schema = test_df.schema
        assert train_schema.names == ["a"], "Only column \"a\" must be present in train DataFrame."
        assert test_schema.names == ["a"], "Only column \"a\" must be present in test DataFrame."

    for (train_df, test_df) in cv.loc(1):
        assert train_df.num_columns == 1, "Only column \"b\" must be present in train DataFrame."
        assert test_df.num_columns == 1, "Only column \"b\" must be present in test DataFrame."
        train_schema = train_df.schema
        test_schema = test_df.schema
        assert train_schema.names == ["b"], "Only column \"b\" must be present in train DataFrame."
        assert test_schema.names == ["b"], "Only column \"b\" must be present in test DataFrame."

    for (train_df, test_df) in cv.loc(["b", "d"]):
        assert train_df.num_columns == 2, "Only columns [\"b\", \"d\"] must be present in train DataFrame."
        assert test_df.num_columns == 2, "Only column [\"b\", \"d\"] must be present in test DataFrame."
        train_schema = train_df.schema
        test_schema = test_df.schema
        assert train_schema.names == ["b", "d"], "Only column [\"b\", \"d\"] must be present in train DataFrame."
        assert test_schema.names == ["b", "d"], "Only column [\"b\", \"d\"] must be present in test DataFrame."

    for (train_df, test_df) in cv.loc([0, 2]):
        assert train_df.num_columns == 2, "Only columns [\"a\", \"c\"] must be present in train DataFrame."
        assert test_df.num_columns == 2, "Only column [\"a\", \"c\"] must be present in test DataFrame."
        train_schema = train_df.schema
        test_schema = test_df.schema
        assert train_schema.names == ["a", "c"], "Only column [\"a\", \"c\"] must be present in train DataFrame."
        assert test_schema.names == ["a", "c"], "Only column [\"a\", \"c\"] must be present in test DataFrame."


def test_cv_null():
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    non_null = df_null.dropna()
    cv = pbn.CrossValidation(df_null)

    for (train_df, test_df), (train_indices, test_indices) in zip(cv, cv.indices()):
        assert non_null.shape[0] == (train_df.num_rows + test_df.num_rows), "CV did not remove null instances correctly."

        nptrain = np.asarray(train_indices)
        nptest = np.asarray(test_indices)
        combination = np.hstack((nptrain, nptest))

        actual_combination = np.sort(np.setdiff1d(np.arange(SIZE), np.asarray(list(set(list(a_null) + list(b_null) + list(c_null) + list(d_null))))))
        
        assert np.all(np.sort(combination) == actual_combination), "Not all the examples are included in the cross validation."
        assert np.all(train_df.to_pandas().to_numpy() == df.iloc[train_indices,:].to_numpy()), \
                                "The CV iterator do not slice the train dataset exactly equal as the CV indices iterator."
        assert np.all(test_df.to_pandas().to_numpy() == df.iloc[test_indices,:].to_numpy()), \
                                "The CV iterator do not slice the test dataset exactly equal as the CV indices iterator."

        assert np.setdiff1d(nptrain, nptest).shape == nptrain.shape, "The train indices includes test indices"
        assert np.setdiff1d(nptest, nptrain).shape == nptest.shape, "The test indices includes train indices"
        assert np.all(np.sort(np.setdiff1d(train_indices, test_indices)) == np.sort(train_indices)), "The train indices includes test indices"
        assert np.all(np.sort(np.setdiff1d(test_indices, train_indices)) == np.sort(test_indices)), "The test indices includes train indices"

    cv_include_null = pbn.CrossValidation(df_null, include_null=True)

    for (train_df, test_df), (train_indices, test_indices) in zip(cv_include_null, cv_include_null.indices()):
        assert (train_df.num_rows + test_df.num_rows) == SIZE, "CV did not remove null instances correctly."

        nptrain = np.asarray(train_indices)
        nptest = np.asarray(test_indices)
        combination = np.hstack((nptrain, nptest))

        train_df_mat = train_df.to_pandas().to_numpy()
        train_indices_mat = df.iloc[train_indices,:].to_numpy()
        test_df_mat = test_df.to_pandas().to_numpy()
        test_indices_mat = df.iloc[test_indices,:].to_numpy()
        
        assert np.all(np.sort(combination) == np.arange(SIZE)), "Not all the examples are included in the cross validation."
        assert np.all(np.isnan(train_df_mat) == np.isnan(train_indices_mat)), \
                                                                "The null values are wrongly specified in the train DataFrame."

        assert np.all(train_df_mat[~np.isnan(train_df_mat)] == train_indices_mat[~np.isnan(train_df_mat)]), \
                                "The CV iterator do not slice the train dataset exactly equal as the CV indices iterator."

        assert np.all(np.isnan(test_df_mat) == np.isnan(test_indices_mat)), \
                                                                "The null values are wrongly specified in the test DataFrame."
        assert np.all(test_df_mat[~np.isnan(test_df_mat)] == test_indices_mat[~np.isnan(test_df_mat)]), \
                                "The CV iterator do not slice the test dataset exactly equal as the CV indices iterator."

        assert np.setdiff1d(nptrain, nptest).shape == nptrain.shape, "The train indices includes test indices"
        assert np.setdiff1d(nptest, nptrain).shape == nptest.shape, "The test indices includes train indices"
        assert np.all(np.sort(np.setdiff1d(train_indices, test_indices)) == np.sort(train_indices)), "The train indices includes test indices"
        assert np.all(np.sort(np.setdiff1d(test_indices, train_indices)) == np.sort(test_indices)), "The test indices includes train indices"


import numpy as np
from scipy.stats import norm
import pybnesian as pbn
import util_test

SIZE = 10000

df = util_test.generate_normal_data(SIZE)

def numpy_local_score(data, variable, evidence):
    if isinstance(variable, str):
        node_data = data.loc[:, [variable] + evidence].dropna()
        variable_data = node_data.loc[:, variable]
        evidence_data = node_data.loc[:, evidence]
    else:
        node_data = data.iloc[:, [variable] + evidence].dropna()
        variable_data = node_data.iloc[:, 0]
        evidence_data = node_data.iloc[:, 1:]

    N = variable_data.shape[0]
    d = evidence_data.shape[1]
    linregress_data = np.column_stack((np.ones(N), evidence_data.to_numpy()))
    (beta, res, _, _) = np.linalg.lstsq(linregress_data, variable_data.to_numpy(), rcond=None)
    var = res / (N - d - 1)

    means = beta[0] + np.sum(beta[1:]*evidence_data, axis=1)
    loglik = norm.logpdf(variable_data, means, np.sqrt(var))

    return loglik.sum() - np.log(N) * 0.5 * (d + 2)

def test_bic_local_score():
    gbn = pbn.GaussianNetwork(['a', 'b', 'c', 'd'], [('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    bic = pbn.BIC(df)
    
    assert np.isclose(bic.local_score(gbn, 'a', []), numpy_local_score(df, 'a', []))
    assert np.isclose(bic.local_score(gbn, 'b', ['a']), numpy_local_score(df, 'b', ['a']))
    assert np.isclose(bic.local_score(gbn, 'c', ['a', 'b']), numpy_local_score(df, 'c', ['a', 'b']))
    assert np.isclose(bic.local_score(gbn, 'd', ['a', 'b', 'c']), numpy_local_score(df, 'd', ['a', 'b', 'c']))
    assert np.isclose(bic.local_score(gbn, 'd', ['a', 'b', 'c']), numpy_local_score(df, 'd', ['b', 'c', 'a']))

    assert bic.local_score(gbn, 'a') == bic.local_score(gbn, 'a', gbn.parents('a'))
    assert bic.local_score(gbn, 'b') == bic.local_score(gbn, 'b', gbn.parents('b'))
    assert bic.local_score(gbn, 'c') == bic.local_score(gbn, 'c', gbn.parents('c'))
    assert bic.local_score(gbn, 'd') == bic.local_score(gbn, 'd', gbn.parents('d'))

def test_bic_local_score_null():
    gbn = pbn.GaussianNetwork(['a', 'b', 'c', 'd'], [('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])

    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan
    
    bic = pbn.BIC(df_null)
    
    assert np.isclose(bic.local_score(gbn, 'a', []), numpy_local_score(df_null, 'a', []))
    assert np.isclose(bic.local_score(gbn, 'b', ['a']), numpy_local_score(df_null, 'b', ['a']))
    assert np.isclose(bic.local_score(gbn, 'c', ['a', 'b']), numpy_local_score(df_null, 'c', ['a', 'b']))
    assert np.isclose(bic.local_score(gbn, 'd', ['a', 'b', 'c']), numpy_local_score(df_null, 'd', ['a', 'b', 'c']))
    assert np.isclose(bic.local_score(gbn, 'd', ['a', 'b', 'c']), numpy_local_score(df_null, 'd', ['b', 'c', 'a']))

    assert bic.local_score(gbn, 'a') == bic.local_score(gbn, 'a', gbn.parents('a'))
    assert bic.local_score(gbn, 'b') == bic.local_score(gbn, 'b', gbn.parents('b'))
    assert bic.local_score(gbn, 'c') == bic.local_score(gbn, 'c', gbn.parents('c'))
    assert bic.local_score(gbn, 'd') == bic.local_score(gbn, 'd', gbn.parents('d'))

def test_bic_score():
    gbn = pbn.GaussianNetwork([('a', 'b'), ('a', 'c'), ('a', 'd'), ('b', 'c'), ('b', 'd'), ('c', 'd')])
    
    bic = pbn.BIC(df)
    
    assert np.isclose(bic.score(gbn), (bic.local_score(gbn, 'a', []) + 
                              bic.local_score(gbn, 'b', ['a']) + 
                              bic.local_score(gbn, 'c', ['a', 'b']) +
                              bic.local_score(gbn, 'd', ['a', 'b', 'c'])))





import numpy as np
import pandas as pd
import pyarrow as pa
import pybnesian as pbn

import pytest
from scipy.stats import norm

import util_test

SIZE = 10000

df = util_test.generate_normal_data(SIZE)

def test_lg_variable():
    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        assert cpd.variable() == variable

def test_lg_evidence():
    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        assert cpd.evidence() == evidence


def fit_numpy(_df, variable, evidence):
    df_na = _df.loc[:, [variable] + evidence].dropna()
    linregress_data = np.column_stack((np.ones(df_na.shape[0]), df_na.loc[:, evidence]))
    (beta, res, _, _) = np.linalg.lstsq(linregress_data, df_na.loc[:, variable], rcond=None)
    
    return beta, res / (df_na.count()[variable] - len(evidence) - 1)

def test_lg_data_type():
    cpd = pbn.LinearGaussianCPD("a", [])
    assert cpd.data_type() == pa.float64()

def test_lg_fit():
    for variable, evidence in [("a", []), ("b", ["a"]), ("c", ["a", "b"]), ("d", ["a", "b", "c"])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        assert not cpd.fitted()
        cpd.fit(df)
        assert cpd.fitted()

        npbeta, npvar = fit_numpy(df, variable, evidence)
        
        assert np.all(np.isclose(npbeta, cpd.beta)), "Wrong beta vector."
        assert np.all(np.isclose(npvar, cpd.variance)), "Wrong variance."

def test_lg_fit_null():
    np.random.seed(0)
    a_null = np.random.randint(0, SIZE, size=100)
    b_null = np.random.randint(0, SIZE, size=100)
    c_null = np.random.randint(0, SIZE, size=100)
    d_null = np.random.randint(0, SIZE, size=100)

    df_null = df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    for variable, evidence in [("a", []), ("b", ["a"]), ("c", ["a", "b"]), ("d", ["a", "b", "c"])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        assert not cpd.fitted()
        cpd.fit(df_null)
        assert cpd.fitted()

        npbeta, npvar = fit_numpy(df_null, variable, evidence)
        
        assert np.all(np.isclose(npbeta, cpd.beta)), "Wrong beta vector."
        assert np.all(np.isclose(npvar, cpd.variance)), "Wrong variance."

def numpy_logpdf(test_df, variable, evidence, beta, variance):
    npdata = test_df.loc[:, evidence].to_numpy()
    means = beta[0] + np.sum(beta[1:]*npdata, axis=1)

    result = np.empty((test_df.shape[0],))

    isnan_vec = np.full((test_df.shape[0],), False, dtype=bool)
    isnan_vec[np.isnan(means)] = True
    isnan_vec[np.isnan(test_df.loc[:, variable].to_numpy())] = True

    result[isnan_vec] = np.nan
    result[~isnan_vec] = norm.logpdf(test_df.loc[:, variable].to_numpy()[~isnan_vec], means[~isnan_vec], np.sqrt(variance))
    return result

def numpy_cdf(test_df, variable, evidence, beta, variance):
    npdata = test_df.loc[:, evidence].to_numpy()
    means = beta[0] + np.sum(beta[1:]*npdata, axis=1)

    result = np.empty((test_df.shape[0],))

    isnan_vec = np.full((test_df.shape[0],), False, dtype=bool)
    isnan_vec[np.isnan(means)] = True
    isnan_vec[np.isnan(test_df.loc[:, variable].to_numpy())] = True

    result[isnan_vec] = np.nan
    result[~isnan_vec] = norm.cdf(test_df.loc[:, variable].to_numpy()[~isnan_vec], means[~isnan_vec], np.sqrt(variance))
    return result

def test_lg_logl():
    test_df = util_test.generate_normal_data(5000)

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        cpd.fit(df)

        beta = cpd.beta
        variance = cpd.variance

        assert np.all(np.isclose(cpd.logl(test_df), numpy_logpdf(test_df, variable, evidence, beta, variance))),\
                     "Wrong logl for LinearGaussianCPD(" + str(variable) + " | " + str(evidence) + ")"

    
    cpd = pbn.LinearGaussianCPD('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.LinearGaussianCPD('d', ['c', 'a', 'b'])
    cpd2.fit(df)

    assert np.all(np.isclose(cpd.logl(test_df), cpd2.logl(test_df))), "The order of the evidence changes the logl() result."

def test_lg_logl_null():
    test_df = util_test.generate_normal_data(5000)

    np.random.seed(0)
    a_null = np.random.randint(0, 5000, size=100)
    b_null = np.random.randint(0, 5000, size=100)
    c_null = np.random.randint(0, 5000, size=100)
    d_null = np.random.randint(0, 5000, size=100)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        cpd.fit(df)

        beta = cpd.beta
        variance = cpd.variance

        assert np.all(np.isclose(
                        cpd.logl(test_df), 
                        numpy_logpdf(test_df, variable, evidence, beta, variance), equal_nan=True)),\
                     "Wrong logl for LinearGaussianCPD(" + str(variable) + " | " + str(evidence) + ") with null values."

    cpd = pbn.LinearGaussianCPD('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.LinearGaussianCPD('d', ['c', 'a', 'b'])
    cpd2.fit(df)

    assert np.all(np.isclose(
                        cpd.logl(test_df), 
                        cpd2.logl(test_df), equal_nan=True)),\
                     "The order of the evidence changes the logl() result."

def test_lg_slogl():
    test_df = util_test.generate_normal_data(5000)

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        cpd.fit(df)

        beta = cpd.beta
        variance = cpd.variance

        assert np.all(np.isclose(cpd.slogl(test_df), np.sum(numpy_logpdf(test_df, variable, evidence, beta, variance)))),\
                     "Wrong slogl for LinearGaussianCPD(" + str(variable) + " | " + str(evidence) + ")"

    cpd = pbn.LinearGaussianCPD('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.LinearGaussianCPD('d', ['c', 'a', 'b'])
    cpd2.fit(df)

    assert np.all(np.isclose(cpd.slogl(test_df), cpd2.slogl(test_df))), "The order of the evidence changes the slogl() result."

def test_lg_slogl_null():
    test_df = util_test.generate_normal_data(5000)

    np.random.seed(0)
    a_null = np.random.randint(0, 5000, size=100)
    b_null = np.random.randint(0, 5000, size=100)
    c_null = np.random.randint(0, 5000, size=100)
    d_null = np.random.randint(0, 5000, size=100)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        cpd.fit(df)

        beta = cpd.beta
        variance = cpd.variance

        assert np.all(np.isclose(cpd.slogl(df_null), np.nansum(numpy_logpdf(df_null, variable, evidence, beta, variance)))),\
                     "Wrong slogl for LinearGaussianCPD(" + str(variable) + " | " + str(evidence) + ") with null values."

    cpd = pbn.LinearGaussianCPD('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.LinearGaussianCPD('d', ['c', 'a', 'b'])
    cpd2.fit(df)

    assert np.all(np.isclose(cpd.slogl(df_null), cpd2.slogl(df_null))), "The order of the evidence changes the slogl() result."

def test_lg_cdf():
    test_df = util_test.generate_normal_data(5000)

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        cpd.fit(df)

        beta = cpd.beta
        variance = cpd.variance

        assert np.all(np.isclose(cpd.cdf(test_df), numpy_cdf(test_df, variable, evidence, beta, variance))),\
                     "Wrong cdf for LinearGaussianCPD(" + str(variable) + " | " + str(evidence) + ")"

    
    cpd = pbn.LinearGaussianCPD('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.LinearGaussianCPD('d', ['c', 'a', 'b'])
    cpd2.fit(df)

    assert np.all(np.isclose(cpd.cdf(test_df), cpd2.cdf(test_df))), "The order of the evidence changes the cdf() result."

def test_lg_cdf_null():
    test_df = util_test.generate_normal_data(5000)

    np.random.seed(0)
    a_null = np.random.randint(0, 5000, size=100)
    b_null = np.random.randint(0, 5000, size=100)
    c_null = np.random.randint(0, 5000, size=100)
    d_null = np.random.randint(0, 5000, size=100)

    df_null = test_df.copy()
    df_null.loc[df_null.index[a_null], 'a'] = np.nan
    df_null.loc[df_null.index[b_null], 'b'] = np.nan
    df_null.loc[df_null.index[c_null], 'c'] = np.nan
    df_null.loc[df_null.index[d_null], 'd'] = np.nan

    for variable, evidence in [('a', []), ('b', ['a']), ('c', ['a', 'b']), ('d', ['a', 'b', 'c'])]:
        cpd = pbn.LinearGaussianCPD(variable, evidence)
        cpd.fit(df)

        beta = cpd.beta
        variance = cpd.variance

        assert np.all(np.isclose(
                        cpd.cdf(df_null), 
                        numpy_cdf(df_null, variable, evidence, beta, variance), equal_nan=True)),\
                     "Wrong cdf for LinearGaussianCPD(" + str(variable) + " | " + str(evidence) + ") with null values."

    cpd = pbn.LinearGaussianCPD('d', ['a', 'b', 'c'])
    cpd.fit(df)
    cpd2 = pbn.LinearGaussianCPD('d', ['c', 'a', 'b'])
    cpd2.fit(df)

    assert np.all(np.isclose(
                        cpd.cdf(df_null), 
                        cpd2.cdf(df_null), equal_nan=True)),\
                     "The order of the evidence changes the cdf() result."

def test_lg_sample():
    SAMPLE_SIZE = 1000

    cpd = pbn.LinearGaussianCPD('a', [])
    cpd.fit(df)
    
    sampled = cpd.sample(SAMPLE_SIZE, None, 0)

    assert sampled.type == pa.float64()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE
        
    cpd = pbn.LinearGaussianCPD('b', ['a'])
    cpd.fit(df)

    sampling_df = pd.DataFrame({'a': np.full((SAMPLE_SIZE,), 3.0)})
    sampled = cpd.sample(SAMPLE_SIZE, sampling_df, 0)

    assert sampled.type == pa.float64()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE
    
    cpd = pbn.LinearGaussianCPD('c', ['a', 'b'])
    cpd.fit(df)

    sampling_df = pd.DataFrame({'a': np.full((SAMPLE_SIZE,), 3.0),
                                'b': np.full((SAMPLE_SIZE,), 7.45)})
    sampled = cpd.sample(SAMPLE_SIZE, sampling_df, 0)

    assert sampled.type == pa.float64()
    assert int(sampled.nbytes / (sampled.type.bit_width / 8)) == SAMPLE_SIZE


import pybnesian as pbn

def test_OperatorTabuSet():
    tabu_set = pbn.OperatorTabuSet()

    assert tabu_set.empty()

    assert not tabu_set.contains(pbn.AddArc("a", "b", 1))    
    tabu_set.insert(pbn.AddArc("a", "b", 2))
    assert not tabu_set.empty()
    assert tabu_set.contains(pbn.AddArc("a", "b", 3))    

    assert not tabu_set.contains(pbn.RemoveArc("b", "c", 4))
    tabu_set.insert(pbn.RemoveArc("b", "c", 5))
    assert tabu_set.contains(pbn.RemoveArc("b", "c", 6))

    assert not tabu_set.contains(pbn.FlipArc("c", "d", 7))
    tabu_set.insert(pbn.RemoveArc("c", "d", 8))
    assert tabu_set.contains(pbn.RemoveArc("c", "d", 9))

    tabu_set.clear()
    assert tabu_set.empty()


